import logging
import numpy as np
import scipy.sparse as spp
from contextlib import contextmanager

from indigo.transforms import Visitor

log = logging.getLogger(__name__)

class Memusage(Visitor):
    """
    Estimates maximum memory needed for evaluating a tree
    during CG.
    """
    def measure(self, node, ncols=1):
        self._seen = set()
        self._current_mem  = [0]
        self._current_cols = [ncols]
        self._max_mem = 0
        self.visit(node)
        return self._max_mem

    @contextmanager
    def _push_mem(self, nbytes):
        self._current_mem.append(nbytes)
        yield
        self._current_mem.pop()

    @contextmanager
    def _push_cols(self, ncols):
        self._current_cols.append(ncols)
        yield
        self._current_cols.pop()
    
    def generic_visit(self, node):
        self._max_mem = max( self._max_mem, sum(self._current_mem) )
        super(Memusage, self).generic_visit(node)

    def visit_Product(self, node):
        ncols = np.prod(self._current_cols)
        nbytes = node._mem_usage(ncols)
        with self._push_mem(nbytes):
            self.generic_visit(node)

    def visit_KronI(self, node):
        with self._push_cols( node._c ):
            self.generic_visit(node)

    def visit_UnscaledFFT(self, node):
        ncols = np.prod(self._current_cols)
        nbytes = node._mem_usage(ncols)
        with self._push_mem(nbytes):
            self.generic_visit(node)

    def visit_DenseMatrix(self, node):
        if id(node) in self._seen:
            return
        self._seen.add(id(node))
        nbytes = node._matrix.nbytes
        self._current_mem[0] += nbytes

    def visit_SpMatrix(self, node):
        if id(node) in self._seen:
            return
        self._seen.add(id(node))
        nrows = node.shape[0]
        rowptr = (nrows+1) * np.dtype('int32').itemsize
        colind =  node.nnz * np.dtype('int32').itemsize
        data   =  node.nnz * node.dtype.itemsize
        nbytes = data + rowptr + colind
        self._current_mem[0] += nbytes



import pytest
import numpy as np
import numpy.testing as npt
import scipy.sparse as spp
from itertools import product

import indigo
from indigo.backends import available_backends

BACKENDS = available_backends()

@pytest.mark.parametrize("backend,L,M,N,K,density",
    list(product( BACKENDS, [3,4], [5,6], [7,8], [1,8,9,17], [1,0.01,0.1,0.5] ))
)
def test_Realize_Product(backend, L, M, N, K, density):
    b = backend()
    A0_h = indigo.util.randM(L, M, density)
    A1_h = indigo.util.randM(M, N, density)
    A0 = b.SpMatrix(A0_h, name='A0')
    A1 = b.SpMatrix(A1_h, name='A1')
    A = A0 * A1
    A = A.realize()

    # forward
    x = b.rand_array((N,K))
    y = b.rand_array((L,K))
    A.eval(y, x)

    x_h = x.to_host()
    y_act = y.to_host()
    y_exp = A0_h @ (A1_h @ x_h)
    npt.assert_allclose(y_act, y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((L,K))
    y = b.rand_array((N,K))
    A.H.eval(y, x)

    x_h = x.to_host()
    y_act = y.to_host()
    y_exp = A1_h.H @ (A0_h.H @ x_h)
    npt.assert_allclose(y_act, y_exp, rtol=1e-5)

    # shape
    assert A.shape == (L,N)
    assert A.H.shape == (N,L)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,M,N,K,density",
    list(product( BACKENDS, [3,4], [5,6], [7,8], [1,0.01,0.1,0.5] ))
)
def test_StoreMatricesInBestOrder(backend, M, N, K, density):
    b = backend()
    A_h = indigo.util.randM(M, K, density)
    A = b.SpMatrix(A_h, name='A')

    from indigo.transforms import StoreMatricesInBestOrder

    AHH = StoreMatricesInBestOrder().visit(A)

    x = b.rand_array((K,N))
    y_act = b.zero_array((M,N), dtype=x.dtype)
    y_exp = b.zero_array((M,N), dtype=x.dtype)

    A.eval(y_exp, x)
    AHH.eval(y_act, x)

    npt.assert_allclose( y_exp.to_host(), y_act.to_host(), rtol=1e-4 )



import pytest
import numpy as np
import scipy.sparse as spp
import numpy.testing as npt
from itertools import product

import indigo
from indigo.backends import available_backends
BACKENDS = available_backends()


@pytest.mark.parametrize("backend,N,K",
    product( BACKENDS, [23,45], [1,2,3] ) )
def test_compat_Multiply(backend, N, K):
    pymr = pytest.importorskip('pymr')
    b = backend()
    x = indigo.util.rand64c(N,K)
    d = indigo.util.rand64c(N,1)

    D0 = pymr.linop.Multiply( (N,K), (N,K), d, dtype=d.dtype )
    D1 = b.Diag(d)

    y_exp = pymr.util.reshape(D0 * pymr.util.vec(x), (N,K))
    y_act = D1 * x

    npt.assert_allclose(y_act, y_exp, rtol=1e-5)


@pytest.mark.parametrize("backend,X,Y,Z,P,K",
    product( BACKENDS, [23,32], [23,32], [23,32], [1,2,3], [1,2,3] ) )
def test_compat_Zpad(backend, X,Y,Z, P, K):
    pymr = pytest.importorskip('pymr')
    b = backend()

    i_shape = (X, Y, Z, K)
    o_shape = (X+2*P, Y+2*P, Z+2*P, K)

    x = indigo.util.rand64c( *i_shape )

    D0 = pymr.linop.Zpad( o_shape, i_shape, dtype=x.dtype )
    D1 = b.Zpad(o_shape[:3], i_shape[:3], dtype=x.dtype)

    x_indigo = np.asfortranarray(x.reshape((-1,K), order='F'))
    x_pmr = pymr.util.vec(x)
    y_exp = D0 * x_pmr
    y_act = D1 * x_indigo

    y_act = y_act.flatten(order='F')
    npt.assert_equal(y_act, y_exp)


@pytest.mark.parametrize("backend,X,Y,Z,K",
    product( BACKENDS, [23,44], [23,44], [23,44], [1,2,3] ) )
def test_compat_UnscaledFFT(backend, X,Y,Z, K):
    pymr = pytest.importorskip('pymr')
    b = backend()

    shape = (X, Y, Z, K)
    x = indigo.util.rand64c( *shape )

    D0 = pymr.linop.FFT(shape, axes=(0,1,2), normalize=False, center=False, dtype=np.dtype('complex64'))
    D1 = b.UnscaledFFT(shape[:3], dtype=np.dtype('complex64'))

    y_exp = D0 * pymr.util.vec(x)
    y_act = D1 * x.reshape((-1,K), order='F')

    y_act = y_act.flatten(order='F')
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,X,Y,Z,K",
    product( BACKENDS, [23,44], [23,44], [23,44], [1,2,3] ) )
def test_compat_UnitaryFFT(backend, X,Y,Z, K):
    pymr = pytest.importorskip('pymr')
    b = backend()

    shape = (X, Y, Z, K)
    x = indigo.util.rand64c( *shape )

    D0 = pymr.linop.FFT(shape, axes=(0,1,2), center=False, dtype=np.dtype('complex64'))
    D1 = b.FFT(shape[:3], dtype=np.dtype('complex64'))

    y_exp = D0 * pymr.util.vec(x)
    y_act = D1 * x.reshape((-1,K), order='F')

    y_act = y_act.flatten(order='F')
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,X,Y,Z,K",
    product( BACKENDS, [23,44], [23,44], [23,44], [1,2,3] ) )
def test_compat_CenteredFFT(backend, X,Y,Z, K):
    pymr = pytest.importorskip('pymr')
    b = backend()

    shape = (X, Y, Z, K)
    x = indigo.util.rand64c( *shape )

    D0 = pymr.linop.FFT(shape, axes=(0,1,2), dtype=np.dtype('complex64'))
    D1 = b.FFTc(shape[:3], dtype=np.dtype('complex64'))

    y_exp = D0 * pymr.util.vec(x)
    y_act = D1 * x.reshape((-1,K), order='F')

    y_act = y_act.flatten(order='F')
    npt.assert_allclose(abs(y_act), abs(y_exp), rtol=1e-2)


@pytest.mark.parametrize("backend,X,Y,Z,RO,PS,K,oversamp,n,width",
    product( BACKENDS, [23,44], [23,44], [23,44], [33,34], [35,36], [1,2], [1.375, 1.43], [64], [3] ) )
def test_compat_NUFFT(backend, X, Y, Z, RO, PS, K, oversamp, n, width):
    pymr = pytest.importorskip('pymr')
    b = backend()

    c_dims  = (X, Y, Z, K)
    nc_dims = (1, RO, PS, K)
    t_dims  = (3, RO, PS)

    x = indigo.util.rand64c(*c_dims)
    traj = indigo.util.rand64c( *t_dims ).real - 0.5
    kwargs = dict(oversamp=oversamp, width=width, n=n, dtype=x.dtype)

    print(nc_dims, c_dims, traj.shape)
    G0 = pymr.linop.NUFFT(nc_dims, c_dims, traj, **kwargs)
    G1 = b.NUFFT(nc_dims[:3], c_dims[:3], traj, **kwargs)

    x_indigo = np.asfortranarray(x.reshape((-1,K), order='F'))
    x_pmr = pymr.util.vec(x)
    y_exp = G0 * x_pmr
    y_act = G1 * x_indigo

    y_act = y_act.reshape(-1, order='F')
    npt.assert_allclose(abs(y_act), abs(y_exp), rtol=1e-2)


@pytest.mark.parametrize("backend,X,Y,Z,RO,PS,K,n,width",
    product( BACKENDS, [23,44], [23,44], [23,44], [33,34], [35,36], [1,2], [64], [3] ) )
def test_compat_Interp(backend, X, Y, Z, RO, PS, K, n, width):
    pymr = pytest.importorskip('pymr')
    b = backend()

    N = (X, Y, Z, K)
    M = (1, RO, PS, K)
    T = (3, RO, PS)

    import scipy.signal as signal
    beta = 0.1234
    kb = signal.kaiser(2 * n + 1, beta)[n:]

    x = indigo.util.rand64c(*N)
    traj = indigo.util.rand64c(*T).real - 0.5

    G0 = pymr.linop.Interp(M, N, traj, width, kb, dtype=x.dtype)
    y_exp = G0 * pymr.util.vec(x)

    G1 = b.Interp(N[:3], traj, width, kb, dtype=x.dtype)
    y_act = G1 * x.reshape((-1,K), order='F')

    y_act = y_act.flatten(order='F')
    npt.assert_allclose( y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,forward,os",
    product( BACKENDS, [True,False], [1.375,1.43] ))
def test_compat_SENSE(backend, forward, os):
    pymr = pytest.importorskip('pymr')
    b = backend()

    X, Y, Z = 12, 13, 14
    RO, PS, C, T = 15, 16, 3, 1

    img = indigo.util.rand64c(X,  Y,  Z, 1, T)
    mps = indigo.util.rand64c(X,  Y,  Z, C, 1)
    ksp = indigo.util.rand64c(1, RO, PS, C, T)
    dcf = indigo.util.rand64c(1, RO, PS, 1, T).real
    coord=indigo.util.rand64c(3, RO, PS, 1, T).real - 0.5

    # pymr
    P = pymr.linop.Multiply(ksp.shape, ksp.shape, dcf, dtype=img.dtype)
    F = pymr.linop.NUFFT(ksp.shape, mps.shape,  coord, dtype=img.dtype)
    S = pymr.linop.Multiply(mps.shape, img.shape, mps, dtype=img.dtype)
    A_pmr = P * F * S

    # indigo
    P1 = b.KronI(C, b.Diag( dcf, name='dcf' ), name='dcf+')
    F1 = b.KronI(C, b.NUFFT(ksp.shape[:3], mps.shape[:3], coord[...,0,0], dtype=img.dtype), name='nufft')
    S1 = b.VStack([b.Diag(mps[:,:,:,c]) for c in range(C)])
    A_indigo = P1 * F1 * S1

    # check
    if forward:
        exp = A_pmr * pymr.util.vec(img)
        act = A_indigo * img.reshape( (-1,1), order='F' )
    else:
        exp = A_pmr.H * pymr.util.vec(ksp)
        act = A_indigo.H * ksp.reshape( (-1,1), order='F' )

    act = act.flatten(order='F')
    npt.assert_allclose(abs(act), abs(exp), rtol=1e-2)

@pytest.mark.parametrize("backend,N",
    product( BACKENDS, [120,130,140] ))
def test_compat_conjgrad(backend, N):
    pymr = pytest.importorskip('pymr')
    b = backend()

    A = indigo.util.randM( N, N, 0.5 )
    A = A.H @ A # make positive definite
    y = indigo.util.rand64c( N )
    x0 = np.zeros( N, dtype=np.complex64 )

    A_pmr = pymr.linop.Matrix( A.toarray(), dtype=A.dtype )
    x_exp = pymr.alg.cg(A_pmr, A.H * y, x0, maxiter=40)

    A_indigo = b.SpMatrix(A)
    b.cg(A_indigo, y, x0, maxiter=40)
    x_act = x0.copy()

    npt.assert_allclose(x_act, x_exp, rtol=1e-6)


@pytest.mark.parametrize("oversamp,width,beta,X,Y,Z",
    product( [1.2,3.4], [3,4], [.1,1.1], [12,13], [14,15], [16,17] ))
def test_compat_rolloff(oversamp, width, beta, X, Y, Z):
    pymr = pytest.importorskip('pymr')
    N = (X,Y,Z)

    from pymr.noncart import rolloff3 as pymr_rolloff3
    from indigo.noncart import rolloff3 as indigo_rolloff3

    r_exp = pymr_rolloff3(oversamp, width, beta, N)
    r_act =  indigo_rolloff3(oversamp, width, beta, N)
    npt.assert_allclose(r_exp, r_act)


@pytest.mark.parametrize("backend,M,N,alpha",
    product( BACKENDS, [120,130,140], [111,222,333], [1e-2, 1e-3] ))
def test_compat_apgd(backend, M, N, alpha):
    pymr = pytest.importorskip('pymr')
    b = backend()

    A = indigo.util.randM( M, N, 0.5 )
    AHA = A.H @ A # make positive definite
    y = indigo.util.rand64c( M )
    AHy = A.H * y
    x0 = np.zeros( N, dtype=np.complex64 )

    A_pmr = pymr.linop.Matrix( AHA.toarray(), dtype=A.dtype )

    # check pymr
    def gradf_pmr(x):
        return A_pmr * x - AHy

    def proxg_pmr(alpha, x0, x1):
        pass

    x_exp = pymr.alg.gd(gradf_pmr, alpha, x0, proxg=proxg_pmr, maxiter=40)

    # check indigo
    x0[:] = 0
    A_indigo = b.SpMatrix(AHA)
    AHy_d = b.copy_array(AHy)

    def gradf_indigo(gf, x):
        # gf = AHA*x - AHy
        A_indigo.eval(gf, x)
        b.axpy(gf, -1, AHy_d)

    def proxg_indigo(alpha, x):
        # x = thresh(x)
        pass

    b.apgd(gradf_indigo, proxg_indigo, alpha, x0, maxiter=40)
    x_act = x0.copy()

    npt.assert_allclose(x_act, x_exp)


@pytest.mark.parametrize("backend,M,N,K",
    product( BACKENDS, [12,34], [23,45], [1] ) )
def test_compat_Matrix(backend, M, N, K):
    pymr = pytest.importorskip('pymr')
    b = backend()
    A = indigo.util.rand64c(M,N)
    x = indigo.util.rand64c(N,K); x_d = b.copy_array(x)
    y = indigo.util.rand64c(M,K); y_d = b.copy_array(y)

    A0 = pymr.linop.Matrix( A, dtype=A.dtype )
    A1 = b.DenseMatrix(A)

    y_exp = pymr.util.reshape(A0 * pymr.util.vec(x), y.shape)
    y_act = A1 * x

    npt.assert_allclose(y_act, y_exp, rtol=1e-5)



import gc
import sys
import time
import logging
from ctypes import *

import numpy as np
from numpy.ctypeslib import ndpointer

from .backend import Backend

log = logging.getLogger(__name__)

cudart   = cdll.LoadLibrary("libcudart.so")
cusparse = cdll.LoadLibrary("libcusparse.so")
cufft    = cdll.LoadLibrary("libcufft.so")
cublas   = cdll.LoadLibrary("libcublas.so")
nvtx     = cdll.LoadLibrary("libnvToolsExt.so")

class c_complex(c_float * 2):
    def __init__(self, a):
        super().__init__()
        self[0] = a.real
        self[1] = a.imag

class CudaBackend(Backend):

    def __init__(self, device_id=0):
        super(CudaBackend, self).__init__()

        self._cublas_handle = self.cublasHandle_t(self)
        self._cusparse_handle = self.cusparseHandle_t(self)
        self._mat_descr = self.cusparseMatDescr_t(self)

        self.cudaSetDevice( device_id )
        cu_device = c_int()
        self.cudaGetDevice( byref(cu_device) )
        log.info("using CUDA device #%d", cu_device.value)

        self._plans = dict()

    class cudaError_t(c_long):
        def _check(self, backend, fn):
            if self.value != 0:
                name = backend.cudaGetErrorName( self ).decode('ascii')
                desc = backend.cudaGetErrorString( self ).decode('ascii')
                log.critical("%s returned exit code %d: %s (%s)",
                    fn.__name__, self.value, name, desc)
                raise RuntimeError

    def wrap(lib):
        def wrapper(fn):
            libfn = getattr(lib, fn.__name__)
            libfn.argtypes = [fn.__annotations__[arg] for arg in fn.__code__.co_varnames]
            libfn.restype  =  fn.__annotations__['return']
            def wrapped(backend, *args, **kwargs):
                res = libfn(*args, **kwargs)
                if hasattr(res, '_check'):
                    res._check(backend, fn)
                return res
            return wrapped
        return wrapper

    @wrap(cudart)
    def cudaSetDevice(device : c_int) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaGetDevice(device : POINTER(c_int)) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaGetErrorName( err : cudaError_t ) -> c_char_p:
        pass

    @wrap(cudart)
    def cudaGetErrorString( err : cudaError_t ) -> c_char_p:
        pass

    @wrap(cudart)
    def cudaGetDeviceCount(count: POINTER(c_int)) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaDeviceSynchronize() -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaMalloc(devPtr : POINTER(c_ulong), size : c_size_t) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaFree(devPtr : c_ulong) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaMemcpy( dst:c_ulong, src:c_ulong, count:c_size_t, kind:c_int) -> cudaError_t:
        pass

    @wrap(cudart)
    def cudaMemset(
        devPtr : c_ulong,
        size   : c_int,
        count  : c_size_t
    ) -> cudaError_t:
        pass

    cudaMemcpy.HostToHost     =   0
    cudaMemcpy.HostToDevice   =   1
    cudaMemcpy.DeviceToHost   =   2
    cudaMemcpy.DeviceToDevice =   3
    cudaMemcpy.Default        =   4

    def barrier(self):
        self.cudaDeviceSynchronize()

    # -----------------------------------------------------------------------
    # Arrays
    # -----------------------------------------------------------------------
    class dndarray(Backend.dndarray):
        def _copy_from(self, arr):
            src, dst = arr.ctypes.data, self._arr
            size, kind = arr.nbytes, CudaBackend.cudaMemcpy.HostToDevice
            self._backend.cudaMemcpy(dst, src, size, kind)

        def _copy_to(self, arr):
            assert arr.flags['F_CONTIGUOUS']
            src, dst = self._arr, arr.ctypes.data
            size, kind = arr.nbytes, CudaBackend.cudaMemcpy.DeviceToHost
            self._backend.cudaMemcpy(dst, src, size, kind)

        def _copy(self, d_arr):
            src, dst = d_arr._arr, self._arr
            size, kind = self.nbytes, CudaBackend.cudaMemcpy.DeviceToDevice
            self._backend.cudaMemcpy(dst, src, size, kind)

        def _malloc(self, shape, dtype):
            align = 256
            self._fullarr = c_ulong()
            self._backend.cudaMalloc( byref(self._fullarr), self.nbytes + align)
            _arr = self._fullarr.value
            while _arr % align != 0:
                _arr += 1
            return c_ulong(_arr)

        def _free(self):
            self._backend.cudaFree( self._fullarr )

        def _zero(self):
            self._backend.cudaMemset( self._arr, 0, self.nbytes )

        def __getitem__(self, slc):
            if isinstance(slc, slice):
                slc = [slc]
            start, shape = [], []
            for s, n in zip(slc, self.shape):
                if isinstance(s, int):
                    s = slice(s, s+1)
                b, e = s.start or 0, s.stop or n
                if b < 0: b = n+b # remove neg begining indices
                if e < 0: e = n+e # remove neg ending indices
                if e < b: e = b   # disallow negative sizes
                if e > n: e = n   # fix over-slices
                start.append(b)
                shape.append(e-b)
            idx = np.ravel_multi_index(start, self.shape, order='F')
            ptr = self._arr.value + idx * np.dtype(self.dtype).itemsize
            ptr = c_ulong(ptr)
            ld = self._leading_dims
            return self._backend.dndarray(self._backend, tuple(shape),
                self.dtype, ld=ld, own=False, data=ptr)

        @staticmethod
        def from_param(obj):
            if not isinstance(obj, CudaBackend.dndarray):
                raise ArgumentError('{} is not a dndarray'.format( type(obj) ))
            return obj._arr


    # -----------------------------------------------------------------------
    # BLAS Routines
    # -----------------------------------------------------------------------

    class cublasHandle_t(c_void_p):
        def __init__(self, backend):
            super().__init__()
            self._backend = backend
            self._backend.cublasCreate_v2( byref(self) )

        def __del__(self):
            self._backend.cublasDestroy_v2( self )


    class cublasStatus_t(c_int):
        def _check(self, backend, fn):
            if self.value != 0:
                log.critical("%s returned exit code %d", fn.__name__, self.value)
                raise RuntimeError

    @wrap(cublas)
    def cublasCreate_v2( handle : POINTER(cublasHandle_t) ) -> cublasStatus_t:
        pass

    @wrap(cublas)
    def cublasDestroy_v2( handle : cublasHandle_t ) -> cublasStatus_t:
        pass

    def axpy(self, y, alpha, x):
        """ y += alpha * x """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        alpha = np.array(alpha, dtype=np.complex64)
        self.cublasCaxpy_v2( self._cublas_handle,
            y.size, alpha, x._arr, 1, y._arr, 1 )

    @wrap(cublas)
    def cublasCaxpy_v2(
        handle : cublasHandle_t,
        n      : c_int,
        alpha  : ndpointer(dtype=np.complex64, ndim=0),
        x      : c_ulong, incx : c_int,
        y      : c_ulong, incy : c_int,
    ) -> cublasStatus_t:
        pass

    def dot(self, x, y):
        """ returns x^T * y """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        dotc = np.array(0, dtype=np.complex64)
        self.cublasCdotc_v2( self._cublas_handle, x.size,
            x._arr, 1, y._arr, 1, dotc)
        return dotc.real

    @wrap(cublas)
    def cublasCdotc_v2(
        handle : cublasHandle_t,
        n      : c_int,
        x      : c_ulong,
        incx   : c_int,
        y      : c_ulong,
        incy   : c_int,
        result : ndpointer(dtype=np.complex64, ndim=0),
    ) -> cublasStatus_t:
        pass

    def norm2(self, x):
        """ returns ||x||_2"""
        assert isinstance(x, self.dndarray)
        result = np.array(0, dtype=np.complex64)
        self.cublasScnrm2_v2( self._cublas_handle, x.size, x._arr, 1, result )
        return result**2

    @wrap(cublas)
    def cublasScnrm2_v2(
        handle : cublasHandle_t,
        n      : c_int,
        x      : c_ulong,
        incx   : c_int,
        result : ndpointer(dtype=np.complex64, ndim=0),
    ) -> cublasStatus_t:
        pass

    def scale(self, x, alpha):
        """ x *= alpha """
        assert isinstance(x, self.dndarray)
        alpha = np.array(alpha, dtype=np.complex64)
        self.cublasCscal_v2( self._cublas_handle, x.size, alpha, x._arr, 1 )

    @wrap(cublas)
    def cublasCscal_v2(
        handle : cublasHandle_t,
        n      : c_int,
        alpha  : ndpointer(dtype=np.complex64, ndim=0),
        x      : c_ulong,
        incx   : c_int,
    ) -> cublasStatus_t:
        pass

    def cgemm(self, y, M, x, alpha, beta, forward):
        """ y = alpha * M * X + beta * Y """
        assert isinstance(x, self.dndarray)
        alpha = np.array(alpha, dtype=np.complex64)
        beta  = np.array( beta, dtype=np.complex64)
        (m, n), k = y.shape, x.shape[0]
        lda = M.shape[0]
        ldb = x.shape[0]
        ldc = y.shape[0]
        if forward:
            transa = CudaBackend.cublasOperator_t.CUBLAS_OP_N
        else:
            transa = CudaBackend.cublasOperator_t.CUBLAS_OP_C
        transb = CudaBackend.cublasOperator_t.CUBLAS_OP_N
        self.cublasCgemm_v2( self._cublas_handle, transa, transb,
            m, n, k, alpha, M, lda, x, ldb, beta, y, ldc )

    class cublasOperator_t(c_uint):
        CUBLAS_OP_N = 0
        CUBLAS_OP_T = 1
        CUBLAS_OP_C = 2

    @wrap(cublas)
    def cublasCgemm_v2(
        handle : cublasHandle_t,
        transa : cublasOperator_t,
        transb : cublasOperator_t,
        m      : c_int,
        n      : c_int,
        k      : c_int,
        alpha  : ndpointer(dtype=np.complex64, ndim=0),
        M      : dndarray,
        lda    : c_int,
        x      : dndarray,
        ldb    : c_int,
        beta   : ndpointer(dtype=np.complex64, ndim=0),
        y      : dndarray,
        ldc    : c_int,
    ) -> cublasStatus_t:
        pass

    # -----------------------------------------------------------------------
    # FFT Routines
    # -----------------------------------------------------------------------

    class cufftHandle_t(c_int):
        def __init__(self, backend):
            super(CudaBackend.cufftHandle_t, self).__init__()
            self._backend = backend
            self._backend.cufftCreate( byref(self) )

        def __del__(self):
            self._backend.cufftDestroy(self)

            
    class cufftType_t(c_int):
        pass

    class cufftResult_t(c_int):
        def _check(self, backend, fn):
            if self.value != 0:
                log.critical("cufft function %s returned error code %d",
                    fn.__name__, self.value)
                raise RuntimeError

    CUFFT_C2C = 0x29
    CUFFT_FORWARD = -1
    CUFFT_INVERSE =  1

    @wrap(cufft)
    def cufftSetAutoAllocation(
        plan : cufftHandle_t,
        auto : c_int,
    ) -> cufftResult_t:
        pass

    @wrap(cufft)
    def cufftSetWorkArea(
        plan : cufftHandle_t,
        area : dndarray,
    ) -> cufftResult_t:
        pass

    @wrap(cufft)
    def cufftMakePlanMany(
        plan    : cufftHandle_t,
        rank    : c_int,
        n       : POINTER(c_int),
        inembed : POINTER(c_int), istride: c_int, idist: c_int,
        onembed : POINTER(c_int), ostride: c_int, odist: c_int,
        typ     : cufftType_t,
        batch   : c_int,
        workSize: POINTER(c_size_t),
    ) -> cufftResult_t:
        pass

    @wrap(cufft)
    def cufftExecC2C(
        plan    : cufftHandle_t,
        idata   : dndarray,
        odata   : dndarray,
        direction : c_int,
    ) -> cufftResult_t:
        pass

    @wrap(cufft)
    def cufftCreate(
        plan : POINTER(cufftHandle_t),
    ) -> cufftResult_t:
        pass

    @wrap(cufft)
    def cufftDestroy(
        plan : cufftHandle_t,
    ) -> cufftResult_t:
        pass

    def _get_or_create_plan(self, x_shape):
        if x_shape not in self._plans:
            N = x_shape[:-1][::-1]
            batch = x_shape[-1]
            dims = (c_int*len(N))(*N)
            plan = CudaBackend.cufftHandle_t(self)
            ws = c_size_t()
            self.cufftSetAutoAllocation(plan, 0)
            self.cufftMakePlanMany(plan, len(dims), dims,
                None, 0, 0, None, 0, 0, CudaBackend.CUFFT_C2C,
                batch, byref(ws))
            self._plans[x_shape] = (plan, ws.value)
        return self._plans[x_shape]

    def _fft_workspace_size(self, x_shape):
        plan, workSize = self._get_or_create_plan(x_shape)
        return workSize

    def fftn(self, y, x):
        plan, workSize = self._get_or_create_plan(x.shape)
        with self.scratch(nbytes=workSize) as tmp:
            self.cufftSetWorkArea(plan, tmp)
            self.cufftExecC2C(plan, x, y, CudaBackend.CUFFT_FORWARD)

    def ifftn(self, y, x):
        plan, workSize = self._get_or_create_plan(x.shape)
        with self.scratch(nbytes=workSize) as tmp:
            self.cufftSetWorkArea(plan, tmp)
            self.cufftExecC2C(plan, x, y, CudaBackend.CUFFT_INVERSE)

    # -----------------------------------------------------------------------
    # Cusparse
    # -----------------------------------------------------------------------
    class cusparseHandle_t(c_void_p):
        def __init__(self, backend):
            super().__init__()
            self._backend = backend
            self._backend.cusparseCreate( byref(self) )

        def __del__(self):
            self._backend.cusparseDestroy(self)


    class cusparseOperation_t(c_int): pass

    class cusparseMatDescr_t(c_void_p):
        def __init__(self, backend):
            super().__init__()
            self._backend = backend
            self._backend.cusparseCreateMatDescr( byref(self) )

        def __del__(self):
            self._backend.cusparseDestroyMatDescr(self)


    class cusparseStatus_t(c_int):
        def _check(self, backend, fn):
            if self.value != 0:
                log.critical("cusparse function %s returned error code %d",
                    fn.__name__, self.value)
                raise RuntimeError


    CUSPARSE_OPERATION_NON_TRANSPOSE       = 0
    CUSPARSE_OPERATION_TRANSPOSE           = 1
    CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE = 2

    @wrap(cusparse)
    def cusparseCreate(
        handle : POINTER(cusparseHandle_t)
    ) -> cusparseStatus_t:
        pass

    @wrap(cusparse)
    def cusparseDestroy(
        handle : cusparseHandle_t
    ) -> cusparseStatus_t:
        pass

    @wrap(cusparse)
    def cusparseCreateMatDescr(
        descr : POINTER(cusparseMatDescr_t)
    ) -> cusparseStatus_t:
        pass

    @wrap(cusparse)
    def cusparseDestroyMatDescr(
        descr : cusparseMatDescr_t
    ) -> cusparseStatus_t:
        pass

    @wrap(cusparse)
    def cusparseCcsrmm(
        handle     : cusparseHandle_t,
        transA     : cusparseOperation_t,
        m          : c_int,
        n          : c_int,
        k          : c_int,
        nnz        : c_int,
        alpha      : c_complex,
        descrA     : cusparseMatDescr_t,
        csrValA    : dndarray,
        csrRowPtrA : dndarray,
        csrColIndA : dndarray,
        B          : dndarray,
        ldb        : c_int,
        beta       : c_complex,
        C          : dndarray,
        ldc        : c_int,
    ) -> cusparseStatus_t:
        pass

    def ccsrmm(self, y, A_shape, A_indx, A_ptr, A_vals, x, alpha, beta, adjoint=False, exwrite=False):
        m, k = A_shape
        n = x.shape[1]
        ldx = x._leading_dims[0]
        ldy = y._leading_dims[0]
        if adjoint:
            trans = self.CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE
        else:
            trans = self.CUSPARSE_OPERATION_NON_TRANSPOSE
        alpha = c_complex(alpha)
        beta  = c_complex(beta)
        self.cusparseCcsrmm( self._cusparse_handle, trans, m, n, k,
            A_vals.size, (alpha), self._mat_descr,
            A_vals, A_ptr, A_indx, x, ldx, (beta), y, ldy
        )



import numpy as np
from ctypes import cdll

from indigo.backends.cuda import CudaBackend
from indigo.backends import _customgpu

class CustomGpuBackend(CudaBackend):

    def ccsrmm(self, Y, A_shape, A_indx, A_ptr, A_vals, X, alpha, beta, adjoint=False, exwrite=False):
        if adjoint and exwrite:
            (M, K), N = A_shape, X.shape[1]
            ldx = X._leading_dims[0]
            ldy = Y._leading_dims[0]
            _customgpu.exw_csrmm(M, N, K, alpha, A_vals._arr.value, A_indx._arr.value, A_ptr._arr.value,
                X._arr.value, ldx, beta, Y._arr.value, ldy)
        else:
            super(CustomGpuBackend, self).ccsrmm(
                Y, A_shape, A_indx, A_ptr, A_vals, X, alpha, beta, adjoint, exwrite=exwrite
            )



import sys
import time
import logging
import io, copy
import itertools
import numpy as np
import scipy.sparse as spp
from ctypes import c_ulong

from indigo.util import profile

log = logging.getLogger(__name__)

class Operator(object):
    def __init__(self, backend, name='', batch=None):
        self._backend = backend
        self._batch = batch
        self._name = name

    def eval(self, y, x, alpha=1, beta=0, forward=True):
        """ y = A * x """
        if x.ndim == 1: x = x.reshape( (x.shape[0], 1) )
        if y.ndim == 1: y = y.reshape( (y.shape[0], 1) )
        M, N = self.shape if forward else tuple(reversed(self.shape))
        if x.shape[0] != N or \
           y.shape[0] != M or \
           x.shape[1] != y.shape[1]:
            raise ValueError("Dimension mismatch: attemping {} = {} * {} ({}, {})".format(
                y.shape, (M,N), x.shape, forward, type(self)))

        batch_size = self._batch or x.shape[1]
        for b in range(0, x.shape[1], batch_size):
          x_slc = x[:,b:b+batch_size]
          y_slc = y[:,b:b+batch_size]
          self._eval(y_slc, x_slc, alpha=alpha, beta=beta, forward=forward)

    @property
    def shape(self):
        raise NotImplemented()

    @property
    def dtype(self):
        raise NotImplemented()

    def __mul__(self, other):
        if isinstance(other, Operator):
            return Product(self._backend, self, other)
        elif isinstance(other, np.ndarray):
            log.warn("using indigow evaluation interface")
            x = other.reshape( (self.shape[1], -1), order='F' )
            x_d = self._backend.copy_array(x)
            y_d = self._backend.zero_array( (self.shape[0],x.shape[1]), dtype=other.dtype )
            self.eval(y_d, x_d)
            return y_d.to_host()
        else:
            raise ValueError("Cannot multiply Operator by %s" % type(other))
            

    @property
    def H(self):
        return Adjoint(self._backend, self, name=self._name+".H")

    def dump(self):
        """
        Returns a textual representation of the operator tree.
        """
        with io.StringIO('w') as f:
            self._dump(file=f, indent=0)
            s = f.getvalue()
        return s

    def _dump(self, file, indent=0):
        name = self._name or 'noname'
        print('{name}, {type}, {shape}, {dtype}'.format(
            name='|   ' * indent + name, type=type(self).__name__,
            shape=self.shape, dtype=self.dtype), file=file)

    def optimize(self, recipe=None):
        from indigo.transforms import Optimize
        return Optimize(recipe).visit(self)

    def memusage(self, ncols=1):
        from indigo.analyses import Memusage
        return Memusage().measure(self, ncols)


class CompositeOperator(Operator):
    def __init__(self, backend, *children, **kwargs):
        super().__init__(backend, **kwargs)
        self._adopt(children)

    @property
    def dtype(self):
        return self._children[0].dtype

    @property
    def child(self):
        assert len(self._children) == 1
        return self._children[0]

    @property
    def children(self):
        return self._children

    @property
    def left_child(self):
        assert len(self._children) == 2
        return self._children[0]

    @property
    def right_child(self):
        assert len(self._children) == 2
        return self._children[1]

    def _adopt(self, children):
        self._children = children

    def _dump(self, file, indent=0):
        s = super()._dump(file, indent)
        for c in self._children:
            c._dump(file, indent+1)

    def realize(self):
        from indigo.transforms import RealizeMatrices
        return RealizeMatrices().visit(self)


class Adjoint(CompositeOperator):
    def __init__(self, backend, children, *args, **kwargs):
        super().__init__(backend, children, *args, **kwargs)

    @property
    def shape(self):
        return tuple(reversed(self.child.shape))

    @property
    def dtype(self):
        return self.child.dtype

    @property
    def H(self):
        return self.child

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        self.child.eval( y, x, alpha, beta, forward=not forward)


class SpMatrix(Operator):
    def __init__(self, backend, M, **kwargs):
        """
        Create a new Sparse Matrix Operator from a concrete sparse matrix.
        """
        super().__init__(backend, **kwargs)
        assert isinstance(M, spp.spmatrix)
        self._matrix = M
        self._matrix_d = None

        self._allow_exwrite = False

    @property
    def dtype(self):
        return self._matrix.dtype

    @property
    def shape(self):
        return self._matrix.shape

    @property
    def nnz(self):
        return self._matrix.nnz

    def _get_or_create_device_matrix(self):
        if self._matrix_d is None:
            M = self._matrix.tocsr()
            M.sort_indices() # cuda requires sorted indictes
            self._matrix_d = self._backend.csr_matrix(self._backend, M, self._name)
            if not self._allow_exwrite:
                log.debug("disallowing exwrite for %s" % self._name)
                self._matrix_d._exwrite = False
            else:
                log.debug("allowing exwrite for %s" % self._name)
        return self._matrix_d

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        M = self._get_or_create_device_matrix()
        if forward:
            nbytes = M.nbytes + x.nbytes*M._col_frac + y.nbytes*2
        else:
            beta_part = 1 if beta == 0 else 2
            col_part = 1 if M._exwrite else 2
            nbytes = M.nbytes + x.nbytes*M._row_frac + y.nbytes*(beta_part+col_part*M._col_frac)
        nflops = 5 * len(self._matrix.data) * x.shape[1]
        with profile("csrmm", nbytes=nbytes, shape=x.shape, forward=forward, nflops=nflops) as p:
            if forward:
                M.forward(y, x, alpha=alpha, beta=beta)
            else:
                M.adjoint(y, x, alpha=alpha, beta=beta)
        profile.ktime += p.duration


class DenseMatrix(Operator):
    def __init__(self, backend, M, **kwargs):
        """
        Create a new Sparse Matrix Operator from a concrete sparse matrix.
        """
        super().__init__(backend, **kwargs)
        assert isinstance(M, np.ndarray)
        assert M.dtype == np.dtype('complex64')
        assert M.flags['F_CONTIGUOUS']
        assert M.ndim == 2
        self._matrix = M
        self._matrix_d = None

    @property
    def dtype(self):
        return self._matrix.dtype

    @property
    def shape(self):
        return self._matrix.shape

    def _get_or_create_device_matrix(self):
        if self._matrix_d is None:
            self._matrix_d = self._backend.copy_array( self._matrix )
        return self._matrix_d

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        M_d = self._get_or_create_device_matrix()
        (m, n), k = M_d.shape, x.shape[1]
        nflops = m * n * k * 5
        with profile("cgemm", nflops=nflops):
            self._backend.cgemm(y, M_d, x, alpha, beta, forward=forward)


class UnscaledFFT(Operator):
    def __init__(self, backend, ft_shape, dtype=np.dtype('complex64'), forward=True, **kwargs):
        super().__init__(backend, **kwargs)
        self._ft_shape = ft_shape
        self._dtype = dtype

    @property
    def shape(self):
        n = np.prod(self._ft_shape)
        return (n,n)

    @property
    def dtype(self):
        return self._dtype

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        assert alpha == 1 and beta == 0
        X = x.reshape( self._ft_shape + (x.shape[1],) )
        Y = y.reshape( self._ft_shape + (x.shape[1],) )

        lens, batch = np.prod(X.shape[:-1]), X.shape[-1]
        nflops = batch * 5 * lens * np.log2(lens)

        if isinstance(X._arr, np.ndarray):
            ptr = X._arr.ctypes.get_data()
        elif isinstance(X._arr, c_ulong):
            ptr = X._arr.value
        else:
            ptr = 0

        align = 1
        while ptr % align == 0:
            align *= 2
        align //= 2

        with profile("fft", nflops=nflops, shape=X.shape, aligned=align) as p:
            if forward:
                self._backend.fftn(Y, X)
            else:
                self._backend.ifftn(Y, X)
        profile.ktime += p.duration

    def _mem_usage(self, ncols):
        ncols = min(ncols, self._batch or ncols)
        ft_shape = self._ft_shape + (ncols,)
        return self._backend._fft_workspace_size(ft_shape)


class KronI(CompositeOperator):
    def __init__(self, backend, c, *args, **kwargs):
        super().__init__(backend, *args, **kwargs)
        self._c = c

    @property
    def shape(self):
        h, w = self.child.shape
        return (self._c * h, self._c * w)

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        cb = self._c * x.shape[1]
        X = x.reshape( (x.size // cb, cb) )
        Y = y.reshape( (y.size // cb, cb) )
        self.child.eval(Y, X, alpha=alpha, beta=beta, forward=forward)


class BlockDiag(CompositeOperator):
    @property
    def shape(self):
        h, w = 0, 0
        for child in self._children:
            h += child.shape[0]
            w += child.shape[1]
        return h, w

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        h_offset, w_offset = 0, 0
        for C in self._children:
            h, w = C.shape if forward else reversed(C.shape)
            slc_x = slice( w_offset, w_offset+w )
            slc_y = slice( h_offset, h_offset+h )
            C.eval( y[slc_y,:], x[slc_x,:], alpha=alpha, beta=beta, forward=forward)
            h_offset += h
            w_offset += w


class VStack(CompositeOperator):
    @property
    def shape(self):
        h, w = 0, 0
        for child in self._children:
            h += child.shape[0]
            w  = child.shape[1]
        return h, w

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        if forward:
            return self._eval_forward(y, x, alpha, beta)
        else:
            return self._eval_adjoint(y, x, alpha, beta)

    def _eval_forward(self, y, x, alpha=1, beta=0):
        h_offset = 0
        for C in self._children:
            h = C.shape[0]
            slc = slice( h_offset, h_offset+h )
            C.eval( y[slc,:], x, alpha=alpha, beta=beta, forward=True)
            h_offset += h

    def _eval_adjoint(self, y, x, alpha=1, beta=0):
        self._backend.scale(y, beta)
        w_offset = 0
        for C in self._children:
            w = C.shape[0]
            slc = slice( w_offset, w_offset+w )
            C.eval( y, x[slc,:], alpha=alpha, beta=1, forward=False)
            w_offset += w

    def _adopt(self, children):
        widths = [child.shape[1] for child in children]
        names = [child._name for child in children]
        if len(set(widths)) > 1:
            raise ValueError("Mismatched widths in VStack: attempting to stack {}".format(
                list(zip(widths, names))))
        super()._adopt(children)


class HStack(CompositeOperator):
    @property
    def shape(self):
        h, w = 0, 0
        for child in self._children:
            h  = child.shape[0]
            w += child.shape[1]
        return h, w

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        if forward:
            return self._eval_forward(y, x, alpha, beta)
        else:
            return self._eval_adjoint(y, x, alpha, beta)

    def _eval_forward(self, y, x, alpha=1, beta=0):
        self._backend.scale(y, beta)
        w_offset = 0
        for C in self._children:
            w = C.shape[1]
            slc = slice( w_offset, w_offset+w )
            C.eval( y, x[slc,:], alpha=alpha, beta=1, forward=True)
            w_offset += w

    def _eval_adjoint(self, y, x, alpha=1, beta=0):
        w_offset = 0
        for C in self._children:
            w = C.shape[1]
            slc = slice( w_offset, w_offset+w )
            C.eval( y[slc,:], x, alpha=alpha, beta=beta, forward=False)
            w_offset += w

    def _adopt(self, children):
        heights = [child.shape[0] for child in children]
        names = [child._name for child in children]
        if len(set(heights)) > 1:
            raise ValueError("Mismatched heights in HStack: attempting to stack {}".format(
                list(zip(heights, names))))
        super()._adopt(children)


class Product(CompositeOperator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._intermediate = None
        self._name = "{}*{}".format(self.left._name, self.right._name)

    @property
    def left(self):
        return self._children[0]

    @property
    def right(self):
        return self._children[1]

    @property
    def shape(self):
        h = self.left.shape[0]
        w = self.right.shape[1]
        return h, w

    def _adopt(self, children):
        L, R = children
        if L.shape[1] != R.shape[0]:
            raise ValueError("Mismatched shapes in Product: attempting {} x {} ({} x {})".format(
                L.shape, R.shape, L._name, R._name))
        super()._adopt(children)

    def _eval(self, y, x, alpha=1, beta=0, forward=True):
        L, R = self._children
        with self._backend.scratch(shape=(R.shape[0],x.shape[1])) as tmp:
            if forward:
                R.eval(tmp, x, alpha=alpha, beta=0, forward=True)
                L.eval(y, tmp, alpha=1,  beta=beta, forward=True)
            else:
                L.eval(tmp, x, alpha=alpha, beta=0, forward=False)
                R.eval(y, tmp, alpha=1,  beta=beta, forward=False)

    def _mem_usage(self, ncols):
        ncols = min(ncols, self._batch or ncols)
        nrows = self._children[1].shape[0]
        return nrows * ncols * self.dtype.itemsize



import pytest
import logging
import numpy as np
import scipy.sparse as spp
from itertools import product

log = logging.getLogger(__name__)

import indigo.util
from indigo.backends import available_backends
BACKENDS = available_backends()

@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_init(backend, n):
    b = backend()
    arr = indigo.util.rand64c(n)
    d_arr = b.copy_array(arr)
    arr2 = d_arr.to_host()
    np.testing.assert_equal(arr, arr2)


@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy_from(backend, n):
    b = backend()
    arr = indigo.util.rand64c(n)
    d_arr = b.zero_array(arr.shape, arr.dtype)
    d_arr.copy_from(arr)
    arr2 = d_arr.to_host()
    np.testing.assert_equal(arr, arr2)


@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy_to(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.copy_array(arr)
    arr2 = np.zeros_like(arr)
    d_arr.copy_to(arr2)
    np.testing.assert_equal(arr, arr2)


@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.copy_array(arr)
    d_arr2 = d_arr.copy()
    d_arr._zero()
    arr2 = d_arr2.to_host()
    np.testing.assert_equal(arr, arr2)

@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy_into(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.copy_array(arr)
    d_arr2 = b.zero_array(arr.shape, arr.dtype)
    d_arr2.copy(d_arr)
    arr2 = d_arr2.to_host()
    np.testing.assert_equal(arr, arr2)

@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_setitem(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.copy_array(arr)
    d_arr2 = b.zero_array(arr.shape, arr.dtype)
    d_arr2[:] = d_arr
    arr2 = d_arr2.to_host()
    np.testing.assert_equal(arr, arr2)


@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy_from_size_mismatch(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.zero_array((n+1,), arr.dtype)
    with pytest.raises(ValueError):
        d_arr.copy_from(arr)


@pytest.mark.parametrize("backend,n", product( BACKENDS, [4,8,129] ))
def test_array_copy_from_dtype_mismatch(backend, n):
    b = backend()
    arr = np.random.rand(n).astype(np.dtype('complex64'))
    d_arr = b.zero_array(arr.shape, np.complex128)
    with pytest.raises(TypeError):
        d_arr.copy_from(arr)


@pytest.mark.parametrize("backend,N,s",
    product( BACKENDS, [23,45], [-2,-1,1,2] )
)
def test_array_slice_1d_front(backend, N, s):
    b = backend()
    arr = np.arange(10)
    d_arr = b.copy_array(arr)
    d_arr_2 = d_arr[:s]
    arr_2 = d_arr_2.to_host()
    np.testing.assert_equal( arr_2, arr[:s] )


@pytest.mark.parametrize("backend,N,s",
    product( BACKENDS, [23,45], [-2,-1,1,2] )
)
def test_array_slice_1d_back(backend, N, s):
    b = backend()
    arr = np.arange(10)
    d_arr = b.copy_array(arr)
    d_arr_2 = d_arr[s:]
    arr_2 = d_arr_2.to_host()
    np.testing.assert_equal( arr_2, arr[s:] )


@pytest.mark.parametrize("backend,batch,x,y,z",
    product( BACKENDS, [1,2,4,8], [23,24,25], [23,24,25], [23,24,25] )
)
def test_fft(backend, batch, x, y, z):
    b = backend()
    N = (z, y, x, batch)
    v = np.random.rand(*N) + 1j*np.random.rand(*N)
    v = np.require(v, dtype=np.dtype('complex64'), requirements='F')
    ax = (0,1,2)
    
    # check forward
    w_exp = np.fft.fftn(v, axes=ax)
    v_d = b.copy_array(v)
    u_d = b.copy_array(v)
    b.fftn(u_d, v_d)
    w_act = u_d.to_host()
    np.testing.assert_allclose(w_act, w_exp, atol=1e-2)

    # check adjoint
    v_exp = np.fft.ifftn(w_act, axes=ax) * (x*y*z)
    v_d = b.copy_array(w_act)
    u_d = b.copy_array(w_act)
    b.ifftn(u_d, v_d)
    v_act = u_d.to_host()
    np.testing.assert_allclose(v_act, v_exp, atol=1e-2)

    # check unitary
    np.testing.assert_allclose(v, v_act / (x*y*z), atol=1e-6)


@pytest.mark.parametrize("backend,M,N,K,density",
    product( BACKENDS, [23,45], [45,23], [1,8,9,17], [0.01,0.1,0.5] )
)
def test_csr_matrix(backend, M, N, K, density):
    b = backend()
    c = np.dtype('complex64')
    A = indigo.util.randM(M, N, density)
    A_d = b.csr_matrix(b, A)

    # forward
    x = (np.random.rand(N,K) + 1j * np.random.rand(N,K))
    x = np.require(x, dtype=c, requirements='F')
    y_exp = A.astype(c) * x
    x_d = b.copy_array(x)
    y_d = b.zero_array(y_exp.shape, x.dtype)
    A_d.forward(y_d, x_d)
    y_act = y_d.to_host()
    np.testing.assert_allclose(y_exp, y_act, atol=1e-5)

    # adjoint
    x = (np.random.rand(M,K) + 1j * np.random.rand(M,K))
    x = np.require(x, dtype=c, requirements='C')
    y_exp = A.H.astype(c) * x
    x_d = b.copy_array(x)
    y_d = b.zero_array(y_exp.shape, x.dtype)
    A_d.adjoint(y_d, x_d)
    y_act = y_d.to_host()
    np.testing.assert_allclose(y_exp, y_act, atol=1e-5)


@pytest.mark.parametrize("backend", BACKENDS)
def test_op_dump(backend):
    b = backend()
    M, N, K, density = 22, 33, 44, 0.50
    A0 = indigo.util.randM(M, N, density)
    A = b.KronI( 4, b.SpMatrix(A0, name='Leaf'), name='Branches' )
    B = A.H * A; B._name = 'Trunk'
    tree = B.dump()
    assert 'Leaf' in tree
    assert 'Trunk' in tree
    assert 'Branches' in tree


@pytest.mark.parametrize("backend,alpha,n,alpha_i",
    product(BACKENDS, [-1.1, -1, -0.1, 0, 0.1, 1, 1.1], [10,23,129], [-2,0,3])
)
def test_blas_scale(backend, alpha, n, alpha_i):
    b = backend()
    x = (np.random.rand(n) + 1j * np.random.rand(n))
    x = np.require(x, dtype=np.dtype('complex64'), requirements='F')
    x_d = b.copy_array(x)

    alpha = alpha + 1j*alpha_i

    y_exp = x * alpha
    b.scale(x_d, alpha)
    y_act = x_d.to_host()

    np.testing.assert_allclose(y_exp, y_act, atol=1e-5)


@pytest.mark.parametrize("backend,device,n",
    product(BACKENDS, [0], [10,23,129])
)
def test_blas_nrm2(backend, device, n):
    b = backend(device)

    x = (np.random.rand(n) + 1j * np.random.rand(n))
    x = np.require(x, dtype=np.dtype('complex64'), requirements='F')
    x_d = b.copy_array(x)

    y_exp = np.linalg.norm(x) ** 2
    y_act = b.norm2(x_d)

    np.testing.assert_allclose(y_exp, y_act, atol=1e-4)


@pytest.mark.parametrize("backend,n",
    product(BACKENDS, [10,23,129,144])
)
def test_blas_dot(backend, n):
    b = backend()
    x = (np.random.rand(n) + 1j * np.random.rand(n))
    y = (np.random.rand(n) + 1j * np.random.rand(n))
    x = np.require(x, dtype=np.dtype('complex64'), requirements='F')
    y = np.require(y, dtype=np.dtype('complex64'), requirements='F')
    x_d = b.copy_array(x)
    y_d = b.copy_array(y)

    y_exp = np.vdot(x, y).real
    y_act = b.dot(x_d, y_d)

    np.testing.assert_allclose(y_exp, y_act, atol=1e-5)


@pytest.mark.parametrize("backend,n,alpha,alpha_i",
    product(BACKENDS, [10,23,129,144], [-2.1, -1.0, -0.1, 0.0, 0.1, 1.0, 1.2], [0,3])
)
def test_blas_axpy(backend, n, alpha, alpha_i):
    b = backend()
    x = (np.random.rand(n) + 1j * np.random.rand(n))
    y = (np.random.rand(n) + 1j * np.random.rand(n))
    x = np.require(x, dtype=np.dtype('complex64'), requirements='F')
    y = np.require(y, dtype=np.dtype('complex64'), requirements='F')
    x_d = b.copy_array(x)
    y_d = b.copy_array(y)

    alpha = alpha + 1j*alpha_i

    y_exp = y + alpha * x
    b.axpy(y_d, alpha, x_d)

    y_act = y_d.to_host()

    np.testing.assert_allclose(y_exp, y_act, atol=1e-6)


@pytest.mark.parametrize("backend,m,n,k,alpha,beta,forward",
    product(BACKENDS, [10,23,129,144],
                      [10,23,129,144],
                      [10,23,129,144],
                      [1,0.5,0.0],
                      [0,0.5,0.0],
                      [True, False])
)
def test_blas_cgemm(backend, m, n, k, alpha, beta, forward):
    b = backend()

    y = indigo.util.rand64c(m,n)
    M = indigo.util.rand64c(m,k)
    x = indigo.util.rand64c(k,n)

    if not forward:
        x, y = y, x
        M_exp = np.conj(M.T)
    else:
        M_exp = M
    y_exp = alpha * M_exp.dot(x) + beta * y

    y_d = b.copy_array(y)
    M_d = b.copy_array(M)
    x_d = b.copy_array(x)
    b.cgemm(y_d, M_d, x_d, alpha, beta, forward=forward)
    y_act = y_d.to_host()

    np.testing.assert_allclose(y_exp, y_act, atol=1e-3)



import numba as nb
import math
import numpy as np
import scipy.sparse as sparse
__all__ = ['interp_mat', 'interp_funs']


@nb.jit(nopython=True, cache=True)
def lin_interp(table, x):
    if x >= 1:
        return 0.0
    n = len(table)
    idx = int(x * (n - 1))
    frac = x * (n - 1) - idx
    return (1.0 - frac) * table[idx] + frac * table[idx + 1]


@nb.jit(nopython=True, cache=True)
def _interp3_mat(m, N, width, table, coord):

    row = np.empty(m * int(2 * width + 1)**3, dtype=np.int64)
    col = np.empty(m * int(2 * width + 1)**3, dtype=np.int64)
    ker = np.empty(m * int(2 * width + 1)**3, dtype=np.float64)

    c = 0
    for i in range(m):

        pos = (N[0] * coord[0, i] + (N[0] // 2),
               N[1] * coord[1, i] + (N[1] // 2),
               N[2] * coord[2, i] + (N[2] // 2))

        start = (math.ceil(pos[0] - width),
                 math.ceil(pos[1] - width),
                 math.ceil(pos[2] - width))
        
        end = (math.floor(pos[0] + width),
               math.floor(pos[1] + width),
               math.floor(pos[2] + width))

        for z in range(start[2], end[2]):

            wz = lin_interp(table, abs(z - pos[2]) / width)
            jz = (z % N[2]) * N[1] * N[0]

            for y in range(start[1], end[1]):

                wy = wz * lin_interp(table, abs(y - pos[1]) / width)
                jy = (y % N[1]) * N[0] + jz

                for x in range(start[0], end[0]):

                    w = wy * lin_interp(table, abs(x - pos[0]) / width)
                    j = (x % N[0]) + jy

                    row[c] = i
                    col[c] = j
                    ker[c] = w

                    c += 1
    return row[:c], col[:c], ker[:c]


def interp_mat(m, N, width, table, coord, backend):

    ndim = coord.shape[0]

    if ndim == 1:
        _interp_mat = _interp1_mat
    elif ndim == 2:
        _interp_mat = _interp2_mat
    elif ndim == 3:
        _interp_mat = _interp3_mat
    else:
        raise ValueError('Number of dimensions can only be 1, 2 or 3, got %r',
                         ndim)

    row, col, ker = _interp_mat(m, N, width, table, coord)
    
    return sparse.coo_matrix((ker, (row, col)),
                             shape=(m, np.prod(N, dtype=np.int)))



import logging
import os, sys, time
from ctypes import *

import numpy as np
from numpy.ctypeslib import ndpointer

from .backend import Backend

log = logging.getLogger(__name__)

# Find MKL library
dll_ext = '.dylib' if sys.platform == 'darwin' else '.so'
libmkl_rt = cdll.LoadLibrary('libmkl_rt' + dll_ext)

class MklBackend(Backend):

    def __init__(self, device_id=0):
        super(MklBackend, self).__init__()
        log.info('mkl_get_version() reports: %s', self.get_version())
        self._fft_descs = dict()

    def wrap(fn):
        libfn = getattr(libmkl_rt, fn.__name__)
        libfn.argtypes = [fn.__annotations__[arg] for arg in fn.__code__.co_varnames]
        libfn.restype  =  fn.__annotations__['return']
        def wrapped(self, *args, **kwargs):
            res = libfn(*args, **kwargs)
            if isinstance(res, c_long) and res.value != 0:
                s = self.DftiErrorMessage( res )
                raise RuntimeError( s.decode('ascii') )
            return res
        return wrapped

    # -----------------------------------------------------------------------
    # Arrays
    # -----------------------------------------------------------------------
    class dndarray(Backend.dndarray):
        _align = 64

        def _copy_from(self, arr):
            self._arr.flat[:] = arr.flat

        def _copy_to(self, arr):
            arr.flat[:] = self._arr.flat

        def _copy(self, d_arr):
            dst = self._arr.reshape(-1, order='F')
            src = d_arr._arr.reshape(-1, order='F')
            dst.flat[:] = src.flat

        def _malloc(self, shape, dtype):
            elems = np.prod(shape) + self._align
            self._arr_orig = arr = np.ndarray(elems, dtype)
            while arr.ctypes.get_data() % self._align != 0:
                arr = arr[1:]
            arr = arr[:np.prod(shape)]
            arr = np.asfortranarray(arr.reshape(shape))
            return arr

        def _free(self):
            del self._arr_orig

        def _zero(self):
            self._arr[:] = 0

        def __getitem__(self, slc):
            ld = self._leading_dims
            d = self._arr.reshape(self.shape, order='F')[slc]
            return self._backend.dndarray( self._backend, d.shape, d.dtype,
                ld=ld, own=False, data=d )

        def to_host(self):
            return self._arr

        @staticmethod
        def from_param(obj):
            if not isinstance(obj, MklBackend.dndarray):
                raise ArgumentError('{} is not a dndarray'.format( type(obj) ))
            return obj._arr.ctypes.get_as_parameter()

        def as_nparray(self):
            return self._arr
    
    @wrap
    def mkl_get_max_threads() -> c_int:
        pass

    def get_max_threads(self):
        return self.mkl_get_max_threads()

    @wrap
    def mkl_get_version_string( buf : c_char_p, length : c_int ) -> c_void_p:
        pass

    def get_version(self):
        buf = create_string_buffer(128)
        self.mkl_get_version_string(buf, len(buf))
        return buf.value.decode('ascii')

    # -----------------------------------------------------------------------
    # BLAS Routines
    # -----------------------------------------------------------------------
    def axpy(self, y, alpha, x):
        """ y += alpha * x """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        alpha = np.array(alpha, dtype=np.complex64)
        self.cblas_caxpy( y.size, alpha, x._arr, 1, y._arr, 1 )

    def dot(self, x, y):
        """ returns x^T * y """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        dotc = np.array(0, dtype=np.complex64)
        self.cblas_cdotc_sub( x.size, x._arr, 1, y._arr, 1, dotc )
        return dotc.real

    def norm2(self, x):
        """ returns ||x||_2"""
        assert isinstance(x, self.dndarray)
        res = self.cblas_scnrm2( x.size, x._arr, 1 )
        return res**2

    def scale(self, x, alpha):
        """ x *= alpha """
        assert isinstance(x, self.dndarray)
        a = np.array(alpha, dtype=np.complex64)
        self.cblas_cscal( x.size, a, x._arr, 1 )

    def cgemm(self, y, M, x, alpha, beta, forward):
        layout = MklBackend.CBlasLayout.ColMajor
        if forward:
            transa = MklBackend.CBlasTranspose.NoTrans
        else:
            transa = MklBackend.CBlasTranspose.ConjTrans
        transb = MklBackend.CBlasTranspose.NoTrans
        (m, n), k = y.shape, x.shape[0]
        alpha = np.array(alpha, dtype=np.complex64)
        beta  = np.array( beta, dtype=np.complex64)
        lda = M.shape[0]
        ldb = x.shape[0]
        ldc = y.shape[0]
        self.cblas_cgemm(
            layout, transa, transb, m, n, k,
            alpha, M, lda, x, ldb, beta, y, ldc
        )

    class CBlasLayout(c_uint):
        RowMajor = 101
        ColMajor = 102

    class CBlasTranspose(c_uint):
        NoTrans   = 111
        Trans     = 112
        ConjTrans = 113

    @wrap
    def cblas_cgemm(
        layout : CBlasLayout,
        transa : CBlasTranspose,
        transb : CBlasTranspose,
        m     : c_int,
        n     : c_int,
        k     : c_int,
        alpha : ndpointer(dtype=np.complex64, ndim=0),
        a     : dndarray,
        lda   : c_int,
        b     : dndarray,
        ldb   : c_int,
        beta  : ndpointer(dtype=np.complex64, ndim=0),
        c     : dndarray,
        ldc   : c_int,
    ) -> c_void_p:
        pass

    @wrap
    def cblas_caxpy(
        n : c_int,
        a : ndpointer(dtype=np.complex64, ndim=0),
        x : ndpointer(dtype=np.complex64),
        incx : c_int,
        y : ndpointer(dtype=np.complex64),
        incy : c_int,
    ) -> c_void_p:
        pass

    @wrap
    def cblas_cdotc_sub(
        n    : c_int,
        x    : ndpointer(dtype=np.complex64),
        incx : c_int,
        y    : ndpointer(dtype=np.complex64),
        incy : c_int,
        dotc : ndpointer(dtype=np.complex64, ndim=0),
    ) -> c_void_p:
        pass

    @wrap
    def cblas_scnrm2(
        n : c_int,
        x    : ndpointer(dtype=np.complex64),
        incx : c_int,
    ) -> c_float:
        pass

    @wrap
    def cblas_cscal(
        n    : c_int,
        a    : ndpointer(dtype=np.complex64, ndim=0),
        x    : ndpointer(dtype=np.complex64),
        incx : c_int,
    ) -> c_void_p:
        pass



    # -----------------------------------------------------------------------
    # FFT Routines
    # -----------------------------------------------------------------------
    class DFTI_DESCRIPTOR_HANDLE(c_void_p):
        pass

    class status_t(c_long):
        pass

    # read these values out of `gcc -E mkl.h | grep DFTI_WHATEVER`
    # they could change so we're living on the edge here
    DFTI_SINGLE = 35
    DFTI_COMPLEX = 32
    DFTI_INPUT_DISTANCE = 14
    DFTI_OUTPUT_DISTANCE = 15
    DFTI_NUMBER_OF_TRANSFORMS = 7
    DFTI_PLACEMENT = 11
    DFTI_INPLACE = 43
    DFTI_NOT_INPLACE = 44

    def _get_or_create_fft_desc(self, x):
        key = (x.shape, x.dtype)
        if key not in self._fft_descs:
            dims, batch = x.shape[:-1][::-1], x.shape[-1]
            ndim = len(dims)
            if ndim == 1:
                lengths = c_long(dims[0])
            else:
                lengths = (c_long*ndim)(*dims)
            desc = self.DFTI_DESCRIPTOR_HANDLE()
            self.DftiCreateDescriptor( byref(desc),
                self.DFTI_SINGLE, self.DFTI_COMPLEX, ndim, lengths )
            self.DftiSetValue( desc, self.DFTI_NUMBER_OF_TRANSFORMS, batch )
            self.DftiSetValue( desc, self.DFTI_PLACEMENT, self.DFTI_NOT_INPLACE )
            self.DftiSetValue( desc, self.DFTI_INPUT_DISTANCE, np.prod(dims) )
            self.DftiSetValue( desc, self.DFTI_OUTPUT_DISTANCE, np.prod(dims) )
            self.DftiCommitDescriptor( desc )
            self._fft_descs[key] = desc
        return self._fft_descs[key]

    def fftn(self, y, x):
        desc = self._get_or_create_fft_desc( x )
        self.DftiComputeForward( desc, x, y )

    def ifftn(self, y, x):
        desc = self._get_or_create_fft_desc( x )
        self.DftiComputeBackward( desc, x, y )

    def __del__(self):
        for desc in self._fft_descs.values():
            self.DftiFreeDescriptor( byref(desc) )

    @wrap
    def DftiErrorMessage(
        status : status_t,
    ) -> c_char_p:
        pass

    @wrap
    def DftiComputeForward(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
        x_in        : dndarray,
        y_out       : dndarray,
    ) -> status_t:
        pass

    @wrap
    def DftiComputeBackward(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
        x_in        : dndarray,
        y_out       : dndarray,
    ) -> status_t:
        pass

    @wrap
    def DftiCommitDescriptor(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
    ) -> status_t:
        pass

    @wrap
    def DftiSetValue(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
        param       : c_uint,
        value       : c_uint,
    ) -> status_t:
        pass

    @wrap
    def DftiCreateDescriptor(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
        precision   : c_uint,
        domain      : c_uint,
        dimension   : c_long,
        #length     : varies. just pass a ctype instance here
    ) -> status_t:
        pass

    @wrap
    def DftiFreeDescriptor(
        desc_handle : DFTI_DESCRIPTOR_HANDLE,
    ) -> status_t:
        pass

    # -----------------------------------------------------------------------
    # CSRMM Routines
    # -----------------------------------------------------------------------
    class csr_matrix(Backend.csr_matrix):
        _index_base = 1

    @wrap
    def mkl_ccsrmm(
        transA   : c_char*1,
        m        : ndpointer(dtype=np.int32,     ndim=0),
        n        : ndpointer(dtype=np.int32,     ndim=0),
        k        : ndpointer(dtype=np.int32,     ndim=0),
        alpha    : ndpointer(dtype=np.dtype('complex64'), ndim=1),
        matdescA : c_char * 6,
        val      : dndarray,
        indx     : dndarray,
        pntrb    : dndarray,
        pntre    : dndarray,
        b        : dndarray,
        ldb      : ndpointer(dtype=np.int32,     ndim=0),
        beta     : ndpointer(dtype=np.dtype('complex64'), ndim=1),
        c        : dndarray,
        ldc      : ndpointer(dtype=np.int32,     ndim=0),
    ) -> c_void_p :
        pass

    @wrap
    def mkl_ccsrmv(
        transA   : c_char*1,
        m        : ndpointer(dtype=np.int32,     ndim=0),
        k        : ndpointer(dtype=np.int32,     ndim=0),
        alpha    : ndpointer(dtype=np.dtype('complex64'), ndim=1),
        matdescA : c_char * 6,
        val      : dndarray,
        indx     : dndarray,
        pntrb    : dndarray,
        pntre    : dndarray,
        b        : dndarray,
        beta     : ndpointer(dtype=np.dtype('complex64'), ndim=1),
        c        : dndarray,
    ) -> c_void_p :
        pass

    def ccsrmm(self, y, A_shape, A_indx, A_ptr, A_vals, x, alpha, beta, adjoint=False, exwrite=False):
        transA = create_string_buffer(1)
        if adjoint:
            transA[0] = b'C'
        else:
            transA[0] = b'N' 
        ldx = np.array(x._leading_dims[0], dtype=np.int32)
        ldy = np.array(y._leading_dims[0], dtype=np.int32)

        A_ptrb = A_ptr[:-1]
        A_ptre = A_ptr[1:]

        m     = np.array(A_shape[0], dtype=np.int32)
        n     = np.array(x.shape[1], dtype=np.int32)
        k     = np.array(A_shape[1], dtype=np.int32)
        alpha = np.array([alpha],    dtype=np.dtype('complex64'))
        beta  = np.array([beta],     dtype=np.dtype('complex64'))

        descrA = create_string_buffer(6)
        descrA[0] = b'G'
        descrA[2] = b'N'
        descrA[3] = b'F'

        if n == 1:
            self.mkl_ccsrmv(transA, m, k, alpha,
                descrA, A_vals, A_indx, A_ptrb, A_ptre,
                x, beta, y)
        else:
            self.mkl_ccsrmm(transA, m, n, k, alpha,
                descrA, A_vals, A_indx, A_ptrb, A_ptre,
                x, ldx, beta, y, ldy)



import logging

log = logging.getLogger(__name__)

def available_backends():
    backends = []

    try:
        from indigo.backends.np import NumpyBackend
        backends.append( NumpyBackend )
    except Exception as e:
        log.warn("couldn't find NUMPY backend")

    try:
        from indigo.backends.mkl import MklBackend
        backends.append( MklBackend )
    except Exception as e:
        log.warn("couldn't find MKL backend")

    try:
        from indigo.backends.cuda import CudaBackend
        backends.append( CudaBackend )
    except Exception as e:
        log.warn("couldn't find CUDA backend")

    try:
        from indigo.backends.customcpu import CustomCpuBackend
        backends.append( CustomCpuBackend )
    except Exception as e:
        log.warn("couldn't find CustomCpu backend")

    try:
        from indigo.backends.customgpu import CustomGpuBackend
        backends.append( CustomGpuBackend )
    except Exception as e:
        log.warn("couldn't find CustomGpu backend")

    return backends

def get_backend(name, **init):
    """
    Instantiates the requested backend.
    """
    if name == 'mkl':
        from indigo.backends.mkl  import MklBackend
        return MklBackend(**init)
    elif name == 'cuda':
        from indigo.backends.cuda import CudaBackend
        return CudaBackend(**init)
    elif name == 'numpy':
        from indigo.backends.np   import NumpyBackend
        return NumpyBackend(**init)
    elif name == 'customcpu':
        from indigo.backends.customcpu import CustomCpuBackend
        return CustomCpuBackend(**init)
    elif name == 'customgpu':
        from indigo.backends.customgpu import CustomGpuBackend
        return CustomGpuBackend(**init)
    else:
        log.error("unrecognized backend: %s", name)



from ctypes import *

import numpy as np
import scipy.sparse as spp
from numpy.ctypeslib import ndpointer

from indigo.backends.backend import Backend

class NumpyBackend(Backend):

    def __init__(self, device_id=0):
        super(NumpyBackend, self).__init__()

    # -----------------------------------------------------------------------
    # Arrays
    # -----------------------------------------------------------------------
    class dndarray(Backend.dndarray):
        def _copy_from(self, arr):
            self._arr.flat[:] = arr.flat

        def _copy_to(self, arr):
            arr.flat[:] = self._arr.flat

        def _copy(self, d_arr):
            dst = self._arr.reshape(-1, order='F')
            src = d_arr._arr.reshape(-1, order='F')
            dst.flat[:] = src.flat

        def _malloc(self, shape, dtype):
            return np.ndarray(shape, dtype, order='F')

        def _free(self):
            del self._arr

        def _zero(self):
            self._arr[:] = 0

        def __getitem__(self, slc):
            d = self._arr.reshape(self.shape, order='F')[slc]
            ld = self._leading_dims
            return self._backend.dndarray( self._backend, d.shape, d.dtype,
                ld=ld, own=False, data=d)

        @staticmethod
        def from_param(obj):
            if not isinstance(obj, NUMPY.dndarray):
                raise ArgumentError('{} is not a dndarray'.format( type(obj) ))
            return obj._arr.ctypes.get_as_parameter()

    # -----------------------------------------------------------------------
    # BLAS Routines
    # -----------------------------------------------------------------------
    def axpy(self, y, alpha, x):
        """ y += alpha * x """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        y._arr += alpha * x._arr.reshape( y.shape, order='F' )

    def dot(self, x, y):
        """ returns x^T * y """
        assert isinstance(x, self.dndarray)
        assert isinstance(y, self.dndarray)
        return np.vdot( x._arr, y._arr ).real

    def norm2(self, x):
        """ returns ||x||_2"""
        assert isinstance(x, self.dndarray)
        return np.linalg.norm(x._arr)**2

    def scale(self, x, alpha):
        """ x *= alpha """
        assert isinstance(x, self.dndarray)
        x._arr *= alpha

    def cgemm(self, y, M, x, alpha, beta, forward):
        x, y, M = x._arr, y._arr, M._arr
        if forward:
            y[:] = alpha * (M @ x) + beta * y
        else:
            MH = np.conj(M.T)
            y[:] = alpha * (MH @ x) + beta * y

    # -----------------------------------------------------------------------
    # FFT Routines
    # -----------------------------------------------------------------------
    def fftn(self, y, x):
        X = x._arr.reshape( x.shape, order='F' )
        Y = y._arr.reshape( y.shape, order='F' )
        ndim = X.ndim-1
        axes = tuple(range(ndim))
        Y[:] = np.fft.fftn(X, axes=axes)

    def ifftn(self, y, x):
        X = x._arr.reshape( x.shape, order='F' )
        Y = y._arr.reshape( y.shape, order='F' )
        ndim = X.ndim-1
        axes = tuple(range(ndim))
        scale = np.prod( X.shape[:ndim] )
        Y[:] = np.fft.ifftn(X, axes=axes) * scale

    # -----------------------------------------------------------------------
    # CSRMM Routine
    # -----------------------------------------------------------------------
    def ccsrmm(self, y, A_shape, A_indx, A_ptr, A_vals, x, alpha, beta, adjoint=False, exwrite=False):
        A = spp.csr_matrix((A_vals._arr, A_indx._arr, A_ptr._arr), shape=A_shape)
        X = x._arr.reshape( x.shape, order='F')
        Y = y._arr.reshape( y.shape, order='F')
        if adjoint:
            Y[:] = alpha * (A.H @ X) + beta * Y
        else:
            Y[:] = alpha * (A @ X) + beta * Y



import time
import logging
import numpy as np
import scipy.sparse as spp
from ctypes import cdll, c_int

log = logging.getLogger(__name__)

def rand64c(*shape, order='F'):
    """
    Constructs a `np.ndarray` of the requested shape and
    populates it with random np.complex64 values.
    """
    r = np.random.rand(*shape).astype(np.float32)
    i = np.random.rand(*shape).astype(np.float32)
    arr = (r + 1j*i).astype(np.complex64)
    if order == 'F':
        arr = np.asfortranarray(arr)
    return arr


def randM(M, N, density):
    """
    Constructs a `scipy.sparse.spmatrix'  of the requested shape and
    density and populates it with random np.complex64 values.
    """
    A_r = spp.random(M, N, density=density, format='csr', dtype=np.float32).astype(np.complex64)
    A_i = spp.random(M, N, density=density, format='csr', dtype=np.float32)
    A = A_r + 1j * A_i
    return A


class profile(object):
    extra = dict()
    ktime = 0

    def __init__(self, event, **kwargs):
        self._event = event
        self._kwargs = kwargs

    def __enter__(self):
        if log.level > logging.DEBUG:
            return

        self._backend.barrier()
        self._start = time.time()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if log.level > logging.DEBUG:
            return

        self._backend.barrier()
        data = dict(
            duration = time.time() - self._start,
            event    = self._event,
        )
        data.update(self.extra)
        data.update(self._kwargs)

        if 'nflops' in data:
            data['gflop_rate'] = data['nflops'] / data['duration'] * 1e-9
            del data['nflops']

        if 'nbytes' in data:
            data['membw_rate'] = data['nbytes'] / data['duration'] * 1e-9
            del data['nbytes']

        kvs = sorted(data.items(), key=lambda kv: kv[0])

        def fmt(k, v):
            if isinstance(v, float):
                return k, "%2.5g" % v
            else:
                return k, repr(v)

        msg = "PROFILE(%s)" % ", ".join("%s=%s" % fmt(*kv) for kv in kvs)
        log.debug(msg)

        self.duration = data['duration']


class Timer(object):
    def __init__(self):
        self._times = []

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, exc_type, exc_value, traceback):
        self._times.append( time.time() - self._start )

    @property
    def median(self):
        return np.median( self._times )

    @property
    def mean(self):
        return np.mean( self._times )

    @property
    def max(self):
        return np.amax( self._times )

    @property
    def min(self):
        return np.amin( self._times )




import copy
import random
import logging
import numpy as np
import scipy.sparse as spp
from itertools import repeat, combinations
from collections import defaultdict

from indigo.operators import (
    CompositeOperator, Product,
    KronI, BlockDiag,
    VStack, SpMatrix,
    Adjoint
)

log = logging.getLogger(__name__)


class Transform(object):
    """
    Visitor class for manipulating operator trees.

    See Also
    --------
    `ast.NodeTransformer`
    """
    def visit(self, node):
        method_name = "visit_%s" % type(node).__name__
        visitor_method = getattr(self, method_name, None)
        if visitor_method:
            node = visitor_method(node)
        else:
            node = self.generic_visit(node)
        return node

    def generic_visit(self, node):
        if isinstance(node, CompositeOperator):
            node._adopt( [self.visit(c) for c in node._children] )
        return node

class Visitor(object):
    """
    Visitor class for traversing operator trees.

    See Also
    --------
    `ast.NodeVisitor`
    """
    def visit(self, node):
        self.generic_visit(node)
        method_name = "visit_%s" % type(node).__name__
        visitor_method = getattr(self, method_name, None)
        if visitor_method:
            visitor_method(node)

    def generic_visit(self, node):
        if isinstance(node, CompositeOperator):
            for child in node._children:
                self.visit(child)


class Optimize(Transform):
    def __init__(self, recipe):
        super(Transform, self).__init__()
        self._recipe = recipe or []

    def visit(self, node):
        steps = [
            #Normalize,
            #TreeTransformations,
            #OperatorTransformations,
            RealizeMatrices,
            #CoalesceAdjoints,
            StoreMatricesInBestOrder,
        ]

        for Step in self._recipe:
            log.info("running optimization step: %s" % Step.__name__)
            node = Step().visit(node)

        # reserve scratch space
        shape = (node.memusage() // node.dtype.itemsize,)
        b = node._backend
        b._scratch = b.empty_array(shape, node.dtype)
        b._scratch_pos = 0

        return node


class Normalize(Transform):
    def visit_Product(self, node):
        """
        Convert nested binary products into a single n-ary product.
        """
        super().generic_visit(node)
        new_kids = []
        for child in node._children:
            if isinstance(child, Product):
                new_kids.extend(child._children)
            else:
                new_kids.append(child)
        node._adopt(new_kids)
        return node


class OperatorTransformations(Transform):
    def ask(self, op):
        return np.random.randint(1, 16)

    def visit_Product(self, node):
        batch = self.ask("batch size <%s>" % node._name)
        node._batch = batch

    def visit_UnscaledFFT(self, node):
        batch = self.ask("batch size <%s>" % node._name)
        node._batch = batch

class TreeTransformations(Transform):
    """
    Manipulates CompositeOperators.
    """
    def visit_Product(self, node):
        node = self.generic_visit(node)
        return node

        if False and self.ask("avoid distributing %s" % node._name):
            return node
            
        new_kids = []
        left = node._children[0]
        for right in node._children[1:]:
            M = self._distribute(left, right)
            if M is not None:
                left = M
            else:
                new_kids.append(left)
                left = right
        new_kids.append(left)
        if len(new_kids) == 1:
            return new_kids[0]
        else:
            node._adopt(new_kids)
            return node

    def visit_BlockDiag(self, node):
        """ BDiag(A*D,B*E,C*F) => BDiag(A,B,C)*BDiag(D,E,F) """
        node = self.generic_visit(node)
        new_kids = []
        if all( isinstance(kid,Product) for kid in node._children ) and \
                self.ask("distribute <%s>" % node._name) and False:
            for grandkids in zip(*[child._children for child in node._children]):
                new_kids.append( BlockDiag(node._backend, *grandkids) )
            return Product( node._backend, *new_kids, name=node._name )
        else:
            return node

    def visit_KronI(self, node):
        """ KronI(A*B) => KronI(A)*KronI(B) """
        node = self.generic_visit(node)
        child = node._children[0]
        if isinstance(child, Product) and \
                self.ask("distribute <%s>" % node._name) and False:
            grandkids = child._children
            new_kids = [KronI(node._backend, node._c, grandkid, name=child._name) for grandkid in grandkids]
            return Product( node._backend, *new_kids, name=node._name )
        else:
            return node

    def visit_VStack(self, node):
        """ VStack( A*D; B*E; C*F ) => BlockDiag(A,B,C) * VStack(D,E,F) """
        node = self.generic_visit(node)
        new_kids = []
        if all( isinstance(kid,Product) for kid in node._children ) and \
                self.ask("distribute <%s>" % node._name) and False:
            for grandkids in zip(*[child._children for child in node._children]):
                new_kids.append( VStack(node._backend, *grandkids) )
            return Product( node._backend, *new_kids, name=node._name )
        else:
            return node

    def visit_Adjoint(self, node):
        """ Adjoint(CompositeOperator(A, B)) => CompositeOperator(Adjoint(B), Adjoint(A)) """
        node = self.generic_visit(node)
        child = node._children[0]
        if isinstance(child, Product) and self.ask("distribute <%s>" % node._name) and False:
            return Product(node._backend, \
                Adjoint(node._backend, child.right, name=child.right._name), \
                Adjoint(node._backend, child.left, name=child.left._name), \
                name=node._name)
        elif isinstance(child, BlockDiag) and self.ask("distribute <%s>" % node._name) and False:
            return node._backend.BlockDiag([Adjoint(node._backend, c, name=c._name) for c in child._children], name=node._name)
        elif isinstance(child, KronI) and self.ask("distribute <%s>" % node._name) and False:
            return node._backend.KronI(child._c, *[Adjoint(node._backend, c, name=c._name) for c in child._children], name=node._name)
        elif isinstance(child, VStack) and self.ask("distribute <%s>" % node._name) and False:
            return node._backend.HStack([Adjoint(node._backend, c, name=c._name) for c in child._children], name=node._name)
        elif isinstance(child, HStack) and self.ask("distribute <%s>" % node._name) and False:
            return node._backend.VStack([Adjoint(node._backend, c, name=c._name) for c in child._children], name=node._name)
        elif isinstance(child, Adjoint) and False:
            return child._children[0]
        return node

    def _distribute(self, left, right):
        """
        Symbolically multiplies and returns LEFT and RIGHT if possible.
        Otherwise returns None. Generally performs conversions of the form:
            OP(A,B,C) * OP(D,E,F) => OP(A*D) * OP(D*E) * OP(C*F)
        """
        b = left._backend

        if isinstance(left, KronI) and isinstance(right, KronI):
            return b.KronI( left._c, left._child * right._child, name=left._name )

        signatures = {
            (b.BlockDiag, b.BlockDiag) : b.BlockDiag,
            (b.BlockDiag, b.KronI    ) : b.BlockDiag,
            (b.BlockDiag, b.VStack   ) : b.VStack,

            (b.KronI, b.BlockDiag) : b.BlockDiag,
            (b.KronI, b.VStack   ) : b.VStack,
          # (b.KronI, b.KronI    ) : b.KronI, # special case above

            (b.VStack, b.BlockDiag) : b.BlockDiag,
            (b.VStack, b.KronI    ) : b.BlockDiag,
        }

        input_type = ( left.__class__, right.__class__ )
        if input_type in signatures:
            operator = signatures[input_type]
            lc = repeat( left._child) if isinstance( left, KronI) else  left._children
            rc = repeat(right._child) if isinstance(right, KronI) else right._children
            return operator([L*R for L,R in zip(lc,rc)], name=left._name)


class RealizeMatrices(Transform):
    """
    Converts CompositeOps into SpMatrix ops if all
    children of the CompositeOp are SpMatrices.
    """
    def visit_Product(self, node):
        """ Product( SpMatrices+ ) => SpMatrix """
        node = self.generic_visit(node)
        left, right = node._children
        if isinstance(left, SpMatrix) and isinstance(right, SpMatrix):
            name = "{}*{}".format(left._name, right._name)
            log.debug('computing %s * %s', left._name, right._name)
            m = left._matrix @ right._matrix
            return SpMatrix( node._backend, m, name=name )
        else:
            return node

    def visit_VStack(self, node):
        """ VStack( SpMatrices ) => SpMatrix """
        node = self.generic_visit(node)
        if all(isinstance(c, SpMatrix) for c in node._children):
            name = "{}+".format(node._children[0]._name)
            dtype = node._children[0].dtype
            log.debug('stacking %s', ', '.join(c._name for c in node._children))
            m = spp.vstack( [c._matrix for c in node._children], dtype=dtype )
            return SpMatrix( node._backend, m, name=name )
        else:
            return node
    
    def visit_HStack(self, node):
        """ HStack( SpMatrices ) => SpMatrix """
        node = self.generic_visit(node)
        if all(isinstance(c, SpMatrix) for c in node._children):
            name = "{}+".format(node._children[0]._name)
            dtype = node._children[0].dtype
            log.debug('stacking %s', ', '.join(c._name for c in node._children))
            m = spp.hstack( [c._matrix for c in node._children], dtype=dtype )
            return SpMatrix( node._backend, m, name=name )
        else:
            return node

    def visit_BlockDiag(self, node):
        """ BlockDiag( SpMatrices ) => SpMatrix """
        node = self.generic_visit(node)
        if all(isinstance(c, SpMatrix) for c in node.children):
            name = "{}+".format(node._children[0]._name)
            dtype = node._children[0].dtype
            m = spp.block_diag( [c._matrix for c in node._children], dtype=dtype )
            return SpMatrix( node._backend, m, name=name )
        else:
            return node

    def visit_KronI(self, node):
        """ KronI(c, SpMatrix) => SpMatrix """
        node = self.generic_visit(node)
        child = node.child
        if isinstance(child, SpMatrix):
            name = "{}+".format(child._name)
            I = spp.identity(node._c, dtype=child.dtype)
            K = spp.kron(I, child._matrix)
            return SpMatrix( node._backend, K, name=name )
        else:
            return node

    def visit_Adjoint(self, node):
        """ Adjoint(M) ==> M.H """
        node = self.generic_visit(node)
        child = node.child
        if isinstance(child, SpMatrix):
            m = child._matrix.getH()
            name = "{}.H".format(child._name)
            return SpMatrix( node._backend, m, name=name )
        else:
            return node

class StoreMatricesInBestOrder(Transform):
    def visit_SpMatrix(self, node):
        """ SpMatrix => Adjoint(SpMatrix.H) """
        if 'interp' in node._name:
            return node
        else:
            M = node._matrix
            return SpMatrix( node._backend, M.getH(), name=node._name ).H


class SetBatchSizeToOne(Visitor):
    def visit(self, node):
        self.generic_visit(node)
        node._batch = 1



import logging
import abc, time
import numpy as np
import scipy.sparse as spp
from contextlib import contextmanager

import indigo.operators as op
from indigo.util import profile

log = logging.getLogger(__name__)

class Backend(object):
    """
    Provides the routines and data structures necessary to implement
    a linear operator chain on different platforms.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, device_id=0):
        profile._backend = self

    class dndarray(object):
        """
        N-dimensional array in device memory.

        Parameters
        ----------
        backend : indigo.backends.Backend
            Backend instance.
        shape : tuple
            Array shape, a la numpy.
        dtype : numpy.dtype
            Datatype.
        ld : tuple
            Shape of array before slicing, used for ldb/ldc values.
        own : bool
            True if this object malloc'ed the underlying memory.
        data : ?
            Handle to underlying memory.
        """
        __metaclass__ = abc.ABCMeta
        _memory = dict()

        def __init__(self, backend, shape, dtype,
                     ld=None, own=True, data=None, name=''):
            assert isinstance(shape, (tuple,list))
            self.dtype = dtype
            self.shape = shape
            self._backend = backend
            self._leading_dims = ld or shape
            self._own = own
            assert isinstance(backend, Backend), type(backend)
            if data is None:
                self._arr = self._malloc(shape, dtype)
                self._memory[ id(self._arr) ] = (name, shape, dtype)
            else:
                self._arr = data

        def reshape(self, new_shape):
            assert -1 not in new_shape
            assert np.prod(new_shape) == self.size
            return self._backend.dndarray( self._backend,
                new_shape, dtype=self.dtype, own=False, data=self._arr)

        @property
        def size(self):
            return np.prod(self.shape)

        @property
        def itemsize(self):
            return self.dtype.itemsize

        @property
        def nbytes(self):
            return self.size * np.dtype(self.dtype).itemsize

        @property
        def ndim(self):
            return len(self.shape)

        def copy_from(self, arr):
            ''' copy from device when both arrays exist '''
            assert isinstance(arr, np.ndarray)
            if self.size != arr.size:
                raise ValueError("size mismatch, expected {} got {}" \
                    .format(self.shape, arr.shape))
            if self.dtype != arr.dtype:
                raise TypeError("dtype mismatch, expected {} got {}" \
                    .format(self.dtype, arr.dtype))
            if not arr.flags['F_CONTIGUOUS']:
                raise TypeError("order mismatch, expected 'F' got {}" \
                    .format(arr.flags['F_CONTIGUOUS']))
            self._copy_from(arr)

        def copy_to(self, arr):
            ''' copy to device when both arrays exist '''
            assert isinstance(arr, np.ndarray)
            if self.size != arr.size:
                raise ValueError("size mismatch, expected {} got {}" \
                    .format(self.shape, arr.shape))
            if self.dtype != arr.dtype:
                raise TypeError("dtype mismatch, expected {} got {}" \
                    .format(self.dtype, arr.dtype))
            self._copy_to(arr)

        def to_host(self):
            ''' copy from device when host array doesnt exist '''
            arr = np.ndarray(self.shape, self.dtype, order='F')
            self.copy_to(arr)
            return arr

        def copy(self, other=None, name=''):
            ''' copy array on device'''
            if other:
                assert isinstance(other, self._backend.dndarray)
                self._copy(other)
            else:
                other = self._backend.zero_array(self.shape, self.dtype, name=name)
                other._copy(self)
                return other

        @classmethod
        def to_device(cls, backend, arr, name=''):
            ''' copy to device when device array doesnt exist '''
            arr_f = np.require(arr, requirements='F')
            d_arr = cls(backend, arr.shape, arr.dtype, name=name)
            d_arr.copy_from(arr_f)
            return d_arr

        def __del__(self):
            """ destructor """
            if self._own and hasattr(self, '_arr'):
                self._memory.pop( id(self._arr) )
                self._free()

        def __setitem__(self, slc, other):
            #FIXME don't ignore slc
            assert not(slc.start or slc.stop), "dndarray setitem cant slice"
            self._copy(other)

        @abc.abstractmethod
        def __getitem__(self, slc):
            """
            Slice notation. Slices must be contiguous in memory. Returns a view.
            """
            raise NotImplementedError()

        @abc.abstractmethod
        def _copy_from(self, arr):
            """ copy HtoD implementation """
            raise NotImplementedError()

        @abc.abstractmethod
        def _copy_to(self, arr):
            """ copy DtoH implementation """
            raise NotImplementedError()

        @abc.abstractmethod
        def _copy(self, arr):
            """ copy DtoD implementation """
            raise NotImplementedError()

        @abc.abstractmethod
        def _malloc(self, shape, dtype):
            """ malloc implementation """
            raise NotImplementedError()

        @abc.abstractmethod
        def _free(self):
            """ malloc implementation """
            raise NotImplementedError()

        @abc.abstractmethod
        def _zero(self):
            """ set to zero """
            raise NotImplementedError()

        @staticmethod
        def from_param(obj):
            """ convert _arr into ctypes object """
            raise NotImplementedError()

    def copy_array(self, arr, name=''):
        return self.dndarray.to_device(self, arr, name=name)

    def zero_array(self, shape, dtype, name=''):
        d_arr = self.empty_array(shape, dtype, name=name)
        d_arr._zero()
        return d_arr

    def empty_array(self, shape, dtype, name=''):
        d_arr = self.dndarray(self, shape, dtype, name=name)
        return d_arr

    def rand_array(self, shape, dtype=np.dtype('complex64'), name=''):
        x = np.random.random(shape) + 1j*np.random.random(shape)
        x = np.require(x, dtype=np.dtype('complex64'), requirements='F')
        x_d = self.copy_array(x, name=name)
        return x_d

    def get_max_threads(self):
        return 1

    def barrier(self):
        pass

    def mem_usage(self):
        nbytes = 0
        log.info("Memory report:")
        table = []
        for name, shape, dtype in self.dndarray._memory.values():
            n = np.prod(shape) * dtype.itemsize
            table.append( (name, n, shape, dtype) )
            nbytes += n
        for name, n, shape, dtype in sorted(table, key=lambda tup: tup[1]):
            if n > 1e6:
                log.info("  %40s: % 3.0f MB, %20s, %15s", name, n/1e6, shape, dtype)
        return nbytes

    @contextmanager
    def scratch(self, shape=None, nbytes=None):
        assert not (shape is not None and nbytes is not None), \
            "Specify either shape or nbytes to backend.scratch()."
        if nbytes is not None:
            shape = (nbytes//np.dtype('complex64').itemsize,)
        size = np.prod(shape)
        if hasattr(self, '_scratch'):
            pos = self._scratch_pos
            total = self._scratch.size
            assert pos + size <= total, "Not enough scratch memory."
            mem = self._scratch[pos:pos+size].reshape(shape)
            self._scratch_pos += size
            yield mem
            self._scratch_pos -= size
        else:
            log.debug("dynamically allocating scratch space in shape %s", shape)
            mem = self.zero_array(shape, dtype=np.complex64)
            yield mem
            del mem

    # -----------------------------------------------------------------------
    # Operator Building Interface 
    # -----------------------------------------------------------------------

    def SpMatrix(self, M, **kwargs):
        """ A := M """
        assert isinstance(M, spp.spmatrix)
        return op.SpMatrix(self, M, **kwargs)

    def DenseMatrix(self, M, **kwargs):
        """ A := M """
        assert isinstance(M, np.ndarray)
        assert M.ndim == 2
        return op.DenseMatrix(self, M, **kwargs)

    def Diag(self, v, **kwargs):
        """ A := diag(v) """
        v = np.require(v, requirements='F')
        if v.ndim > 1:
            v = v.flatten(order='A')
        dtype = kwargs.get('dtype', np.dtype('complex64'))
        M = spp.diags( v, offsets=0 ).astype(dtype)
        return self.SpMatrix(M, **kwargs)

    def Eye(self, n, dtype=np.dtype('complex64'), **kwargs):
        """ A := Eye(n) """
        M = spp.eye(n, dtype=dtype)
        return self.SpMatrix(M, **kwargs)

    def KronI(self, c, B, **kwargs):
        """ C := I_c (KRON) B """
        return op.KronI(self, c, B, **kwargs)

    def BlockDiag(self, Ms, **kwargs):
        return op.BlockDiag(self, *Ms, **kwargs)

    def VStack(self, Ms, **kwargs):
        return op.VStack(self, *Ms, **kwargs)

    def HStack (self, Ms, **kwargs):
        return op.HStack(self, *Ms, **kwargs)

    def UnscaledFFT(self, shape, dtype, **kwargs):
        """ A := FFT{ . } """
        return op.UnscaledFFT(self, shape, dtype, **kwargs)

    def CopyIn(self, shape, dtype, **kwargs):
        return op.CopyIn(self, shape, dtype)

    def CopyOut(self, shape, dtype, **kwargs):
        return op.CopyOut(self, shape, dtype)

    def FFT(self, shape, dtype, **kwargs):
        """ Unitary FFT """
        n = np.prod(shape)
        s = np.ones(n, order='F', dtype=dtype) / np.sqrt(n)
        S = self.Diag(s, name='scale')
        F = self.UnscaledFFT(shape, dtype, **kwargs)
        return S*F

    def FFTc(self, ft_shape, dtype, **kwargs):
        """ Centered, Unitary FFT """
        mod_slice = [ slice(d) for d in ft_shape ]
        idx = np.mgrid[mod_slice]
        mod = 0
        for i in range(len(ft_shape)):
            c = ft_shape[i] // 2
            mod += (idx[i] - c / 2.0) * (c / ft_shape[i])
        mod = np.exp(1j * 2.0 * np.pi * mod).astype(dtype)
        M = self.Diag(mod, name='mod')
        F = self.FFT(ft_shape, dtype=dtype, **kwargs)
        return M*F*M

    def Zpad(self, M, N, dtype=np.dtype('complex64'), **kwargs):
        slc = []
        for m, n in zip(M, N):
            slc += [slice(m // 2 + int(np.ceil(-n / 2)),
                          m // 2 + int(np.ceil( n / 2))), ]
        x = np.arange( np.prod(M), dtype=int ).reshape(M, order='F')
        rows = x[slc].flatten(order='F')
        cols = np.arange(rows.size)
        ones = np.ones_like(cols)
        shape = np.prod(M), np.prod(N)
        M = spp.coo_matrix( (ones, (rows,cols)), shape=shape, dtype=dtype )
        return self.SpMatrix(M, **kwargs)

    def Interp(self, N, coord, width, table, dtype=np.dtype('complex64'), **kwargs):
        assert len(N) == 3
        ndim  = coord.shape[0]
        npts = np.prod( coord.shape[1:] )
        coord = coord.reshape((ndim,-1), order='F')

        from indigo.interp import interp_mat
        M = interp_mat(npts, N, width, table, coord, 1).astype(dtype)

        return self.SpMatrix(M, **kwargs)

    def NUFFT(self, M, N, coord, width=3, n=128, oversamp=None, dtype=np.dtype('complex64'), **kwargs):
        assert len(M) == 3
        assert len(N) == 3
        assert M[1:] == coord.shape[1:]

        # target 448 x 270 x 640
        #   448 x 270 x 640   mkl-batch: 170.83 ms, 237.51 gflop/s  back-to-back: 121.76 ms, 333.23 gflop/s
        #   1.45  1.30  1.33
        #   432 x 280 x 640   mkl-batch: 183.85 ms  220.7 gflop/s   back-to-back: 149.62 ms  271.19 gflop/s
        #   1.40  1.35  1.33
        #   432 x 270 x 640   mkl-batch: 168.62 ms  231.57 gflop/s  back-to-back: 118.31 ms  330.05 gflop/s
        #   1.40  1.30  1.33

        if isinstance(oversamp, tuple):
            omin = min(oversamp)
        else:
            omin = oversamp
            oversamp = (omin, omin, omin)

        import scipy.signal as signal
        from indigo.noncart import rolloff3
        ndim  = coord.shape[0]
        npts  = np.prod( coord.shape[1:] )

        oN = list(N)
        for i in range(3):
            oN[i] *= oversamp[i]
        oN = tuple(int(on) for on in oN)

        Z = self.Zpad(oN, N, dtype=dtype, name='zpad')
        F = self.FFTc(oN, dtype=dtype, name='fft')

        beta = np.pi * np.sqrt(((width * 2. / omin) * (omin- 0.5)) ** 2 - 0.8)
        kb = signal.kaiser(2 * n + 1, beta)[n:]
        G = self.Interp(oN, coord, width, kb, dtype=np.float32, name='interp')

        r = rolloff3(omin, width, beta, N)
        R = self.Diag(r, name='apod')

        return G*F*Z*R

    # -----------------------------------------------------------------------
    # BLAS Routines
    # -----------------------------------------------------------------------
    def axpy(self, y, alpha, x):
        """ y += alpha * x """
        raise NotImplementedError()

    def dot(self, x, y):
        """ returns x^T * y """
        raise NotImplementedError()

    def norm2(self, x):
        """ returns ||x||_2"""
        raise NotImplementedError()

    def scale(self, x, alpha):
        """ x *= alpha """
        raise NotImplementedError()

    def pdot(self, x, y, comm):
        xHy = self.dot(x, y)
        if comm is not None:
            xHy = comm.allreduce( xHy )
        return xHy

    def pnorm2(self, x, comm):
        xTx = self.norm2(x)
        if comm is not None:
            xTx = comm.allreduce( xTx )
        return xTx

    def cgemm(self, y, M, x, alpha, beta, forward):
        """
        Peform a dense matrix-matrix multiplication.
        """
        raise NotImplementedError()

    # -----------------------------------------------------------------------
    # FFT Routines
    # -----------------------------------------------------------------------

    @abc.abstractmethod
    def fftn(self, y, x):
        """
        Peform an unscaled multidimensional forward FFT on x.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def ifftn(self, y, x):
        """
        Peform an unscaled multidimensional inverse FFT on x.
        """
        raise NotImplementedError()

    def _fft_workspace_size(self, x_shape):
        return 0

    @abc.abstractmethod
    def ccsrmm(self, y, A_shape, A_indx, A_ptr, A_vals, x, alpha=1, beta=0, adjoint=False, exwrite=False):
        """
        Computes Y[:] = A * X.
        """
        raise NotImplementedError()


    class csr_matrix(object):
        """
        A device-resident sparse matrix in CSR format.
        """
        _index_base = 0

        def __init__(self, backend, A, name='mat'):
            """
            Create a matrix from the given `scipy.sparse.sppmatrix`.
            """
            if not isinstance(A, spp.csr_matrix):
                A = A.tocsr()
            A = self._type_correct(A)
            self._backend = backend
            self.rowPtrs = backend.copy_array(A.indptr + self._index_base, name=name+".rowPtrs")
            self.colInds = backend.copy_array(A.indices + self._index_base, name=name+".colInds")
            self.values  = backend.copy_array(A.data, name=name+".data")
            self.shape = A.shape
            self.dtype = A.dtype

            # fraction of nonzero rows/columns
            from indigo.backends._customcpu import inspect
            nzrow, nzcol, self._exwrite = inspect(A.shape[0], A.shape[1], A.indices, A.indptr)
            self._row_frac = nzrow / A.shape[0]
            self._col_frac = nzcol / A.shape[1]
            log.debug("matrix %s has %2d%% nonzero rows and %2d%% nonzero columns",
                name, 100*self._row_frac, 100*self._col_frac)
            log.debug("matrix %s supports exwrite: %s", name, self._exwrite)

        def forward(self, y, x, alpha=1, beta=0):
            """ y[:] = A * x """
            self._backend.ccsrmm(y,
                self.shape, self.colInds, self.rowPtrs, self.values,
                x, alpha=alpha, beta=beta, adjoint=False, exwrite=True)

        def adjoint(self, y, x, alpha=1, beta=0):
            """ y[:] = A.H * x """
            self._backend.ccsrmm(y,
                self.shape, self.colInds, self.rowPtrs, self.values,
                x, alpha=alpha, beta=beta, adjoint=True, exwrite=self._exwrite)

        @property
        def nbytes(self):
            return self.rowPtrs.nbytes + self.colInds.nbytes + self.values.nbytes

        @property
        def nnz(self):
            return self.values.size

        def _type_correct(self, A):
            return A.astype(np.complex64)

    # -----------------------------------------------------------------------
    # Algorithms
    # -----------------------------------------------------------------------

    def cg(self, A, b_h, x_h, lamda=0.0, tol=1e-10, maxiter=100, team=None):
        """
        Conjugate gradient. Solves for A x = b, where A is positive semi-definite.

        Parameters
        ----------
        A : function to perform A(x)
        y : 1D array
        x : 1D array, initial solution
        maxiter : int, optional
        {IterPrint, IterPlot, IterWrite, IterCompare}
        """
        Ap = self.zero_array( (A.shape[0],), x_h.dtype, name='Ap' )
        x = self.copy_array( x_h, name='x' )
        b = self.copy_array( b_h, name='b' )

        # r = b - A(x) - lamda * x
        r = b
        A.eval(Ap, x)

        self.axpy(r, -1, Ap)
        self.axpy(r, -lamda, x)

        p = r.copy(name='p')
        rr = self.pnorm2(r, team)
        r0 = rr

        for it in range(maxiter):
            profile.extra['it'] = it
            profile.ktime = 0
            with profile("iter"):
                A.eval(Ap, p)
                self.axpy(Ap, lamda, p)
                alpha = rr / self.pdot(p, Ap, team)
                self.axpy(x, alpha, p)
                self.axpy(r, -alpha, Ap)

                r2 = self.pnorm2(r, team)
                beta = r2 / rr
                self.scale(p, beta)
                self.axpy(p, 1, r)
                rr = r2

                resid = np.sqrt(rr / r0)
                log.info("iter %d, residual %g", it, resid.real)

                log.info("ktime =%f=" % profile.ktime)

                if resid < tol:
                    log.info("cg reached tolerance")
                    break
        else:
            log.info("cg reached maxiter")
        x.copy_to(x_h)

    def apgd(self, gradf, proxg, alpha, x_h, maxiter=100, team=None):
        '''Accelerated proximal gradient descent.
        Solves for min_x f(x) + g(x)

        Parameters
        ----------
        gradf : Gradient of f
        proxg : Proximal of g
        alpha : Step size
        x0 : 1D array, initial solution
        maxiter : int, optional
        '''
        x = self.copy_array(x_h)
        z = x.copy()
        o = x.copy()
        gf = x.copy()

        t = 1.0
        for it in range(maxiter):
            profile.extra['it'] = it
            with profile("iter"):

                o.copy(x)
                s = t

                gradf(gf, z)
                self.axpy(x, -alpha, gf)
                proxg(alpha, x)

                t = (1.0 + (1.0 + 4.0 * t**2)**0.5) / 2.0

                # z = x + (s-1)/t * (x-o)
                z.copy(x)                # z = x
                self.axpy(z, (s-1)/t, x) # z += (s-1)/t*x
                self.axpy(z, (1-s)/t, o) # z += (1-s)/t*o

                r2 = self.pnorm2(x, team)
                log.info("iter %d, residual %g", it, r2.real)
        x.copy_to(x_h)



import pytest
import numpy as np
import scipy.sparse as spp
import numpy.testing as npt
from itertools import product

import indigo
from indigo.operators import Product
from indigo.backends import available_backends
BACKENDS = available_backends()

@pytest.mark.parametrize("backend,L,M,N,K,density,b0,b1,b2",
    product( BACKENDS, [3,4], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1],
        [None, 1, 2, 10], [None, 1, 2], [None, 1, 2]
))
def test_Memusage_Product(backend, L, M, N, K, density, b0, b1, b2):
    b = backend()
    A0_h = indigo.util.randM(L, M, density)
    A1_h = indigo.util.randM(M, N, density)
    A0 = b.SpMatrix(A0_h, name='A0', batch=b0)
    A1 = b.SpMatrix(A1_h, name='A1', batch=b1)
    A = Product(b, A0, A1, batch=b2)

    # x = b.rand_array((N,K))
    # y = b.rand_array((L,K))

    a0_nbytes = A0_h.data.nbytes + A0_h.indices.nbytes + A0_h.indptr.nbytes
    a1_nbytes = A1_h.data.nbytes + A1_h.indices.nbytes + A1_h.indptr.nbytes

    # forward
    x_nbytes  = N * K * A.dtype.itemsize
    bs = min(K, b2 or K)
    tmp_nbytes = A0.shape[1] * bs * A0.dtype.itemsize
    nbytes_exp = a0_nbytes + a1_nbytes + tmp_nbytes
    nbytes_act = A.memusage(ncols=K)
    assert nbytes_exp == nbytes_act

    # adjoint
    x_nbytes  = L * K * A.dtype.itemsize
    tmp_nbytes = A0.H.shape[0] * bs * A.dtype.itemsize
    nbytes_exp = a0_nbytes + a1_nbytes + tmp_nbytes
    nbytes_act = A.H.memusage(ncols=K)
    assert nbytes_exp == nbytes_act



import numpy as np
import numexpr as ne


def ftkb(beta, x):
    pi = np.pi
    a = ne.evaluate('sqrt(beta ** 2 - (pi * x) ** 2)')

    ret = np.empty(a.shape, dtype=a.dtype)
    ret[a == 0.0] = 1.0
    b = a[a != 0.0]
    ret[a != 0.0] = ne.evaluate('sinh(b) / b')

    return ret


def rolloff1(oversamp, width, beta, N):

    x = np.mgrid[:N[0]]

    return ftkb(beta, 0.0) / ftkb(beta, (x - N[0] // 2) / N[0] * width * 2.0 / oversamp)


def rolloff2(oversamp, width, beta, N):

    x, y = np.mgrid[:N[0], :N[1]]

    return ftkb(beta, 0.0)**2 / (ftkb(beta, (x - N[0] // 2) / N[0] * width * 2.0 / oversamp) *
                                 ftkb(beta, (y - N[1] // 2) / N[1] * width * 2.0 / oversamp))


def rolloff3(oversamp, width, beta, N):

    x, y, z = np.mgrid[:N[0], :N[1], :N[2]]

    return ftkb(beta, 0.0)**3 / (ftkb(beta, (x - N[0] // 2) / N[0] * width * 2.0 / oversamp) *
                                 ftkb(beta, (y - N[1] // 2) / N[1] * width * 2.0 / oversamp) *
                                 ftkb(beta, (z - N[2] // 2) / N[2] * width * 2.0 / oversamp))



import os
from distutils.core import setup, Extension
import numpy as np

exts = []

customcpu_backend = Extension('indigo.backends._customcpu',
    sources = ['indigo/backends/_customcpu.c'],
    include_dirs=[np.get_include()],
    extra_compile_args = ['-std=c11', '-fopenmp', '-m64', '-O3', '-Wno-unknown-pragmas'],
    extra_link_args=['-fopenmp', '-mavx'],
)
exts.append(customcpu_backend)

if os.path.exists( 'indigo/backends/libgpu.a' ):
    customgpu_backend = Extension('indigo.backends._customgpu',
        sources = ['indigo/backends/_customgpu.c'],
        include_dirs=[np.get_include()],
        extra_compile_args = ['-std=c11'],
        extra_link_args=['indigo/backends/libgpu.a', '-L/usr/local/cuda/lib64', '-lcudart', '-lcublas', '-lstdc++'],
    )
    exts.append(customgpu_backend)

setup(
    name = 'indigo',
    packages = ['indigo', 'indigo.backends'],
    version = '1.0.0',
    author = 'Michael Driscoll',
    author_email = 'mbdriscoll@gmail.com',
    license = 'BSD',
    url = 'https://mbdriscoll.github.io/indigo/',
    download_url = 'https://github.com/mbdriscoll/indigo/archive/master.zip',
    ext_modules = exts,
)



import numpy as np
from ctypes import cdll

from indigo.backends.mkl import MklBackend
from indigo.backends import _customcpu

class CustomCpuBackend(MklBackend):

    class csr_matrix(MklBackend.csr_matrix):
        _index_base = 0

        def _type_correct(self, A):
            return A

    def ccsrmm(self, Y, A_shape, A_indx, A_ptr, A_vals, X, alpha, beta, adjoint=False, exwrite=False):
        ldx = X._leading_dims[0]
        ldy = Y._leading_dims[0]
        (M, K), N = A_shape, X.shape[1]
        _customcpu.csrmm(adjoint, M, N, K, alpha,
            A_vals._arr, A_indx._arr, A_ptr._arr,
            X._arr, ldx, beta, Y._arr, ldy, exwrite)



import pytest
import numpy as np
import scipy.sparse as spp
import numpy.testing as npt
from itertools import product

import indigo
from indigo.backends import available_backends
BACKENDS = available_backends()

@pytest.mark.parametrize("backend,M,N,K,density,alpha,beta",
    product( BACKENDS, [23,45], [45,23], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1] ))
def test_SpMatrix(backend, M, N, K, density, alpha, beta):
    b = backend()
    A_h = indigo.util.randM(M, N, density)
    A = b.SpMatrix(A_h)

    # forward
    x = b.rand_array((N,K))
    y = b.rand_array((M,K))
    y_exp = beta * y.to_host() + alpha * A_h * x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((M,K))
    y = b.rand_array((N,K))
    y_exp = beta * y.to_host() + alpha * A_h.H * x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # double adjoint
    x = b.rand_array((N,K))
    y = b.rand_array((M,K))
    y_exp = beta * y.to_host() + alpha * A_h.H.H * x.to_host()
    A.H.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # triple adjoint
    x = b.rand_array((M,K))
    y = b.rand_array((N,K))
    y_exp = beta * y.to_host() + alpha * A_h.H.H.H * x.to_host()
    A.H.H.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (M,N)
    assert A.H.shape == (N,M)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,L,M,N,K,density,alpha,beta",
    product( BACKENDS, [3,4], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1] ))
def test_Product(backend, L, M, N, K, density, alpha, beta):
    b = backend()
    A0_h = indigo.util.randM(L, M, density)
    A1_h = indigo.util.randM(M, N, density)
    A0 = b.SpMatrix(A0_h, name='A0')
    A1 = b.SpMatrix(A1_h, name='A1')
    A = A0 * A1

    # forward
    x = b.rand_array((N,K))
    y = b.rand_array((L,K))
    y_exp = beta * y.to_host() + alpha * A0_h @ (A1_h @ x.to_host())
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((L,K))
    y = b.rand_array((N,K))
    y_exp = beta * y.to_host() + alpha * A1_h.H @ (A0_h.H @ x.to_host())
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (L,N)
    assert A.H.shape == (N,L)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,stack,M,N,K,density,alpha,beta",
    product( BACKENDS, [1,2,3], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,1,0.5] ))
def test_VStack(backend, stack, M, N, K, density, alpha, beta):
    b = backend()
    mats_h = [indigo.util.randM(M,N,density) for i in range(stack)]
    A_h = spp.vstack(mats_h)

    mats_d = [b.SpMatrix(m) for m in mats_h]
    A = b.VStack(mats_d)

    # forward
    x = b.rand_array((A.shape[1],K))
    y = b.rand_array((A.shape[0],K))
    y_exp = beta * y.to_host() + alpha * A_h @ x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((A.shape[0],K))
    y = b.rand_array((A.shape[1],K))
    y_exp = beta * y.to_host() + alpha * A_h.H @ x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (M*stack,N)
    assert A.H.shape == (N,M*stack)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,stack,M,N,K,density,alpha,beta",
    product( BACKENDS, [1,2,3], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1] ))
def test_HStack(backend, stack, M, N, K, density, alpha, beta):
    b = backend()
    mats_h = [indigo.util.randM(M,N,density) for i in range(stack)]
    A_h = spp.hstack(mats_h)

    mats_d = [b.SpMatrix(m) for m in mats_h]
    A = b.HStack(mats_d)

    # forward
    x = b.rand_array((A.shape[1],K))
    y = b.rand_array((A.shape[0],K))
    y_exp = beta * y.to_host() + alpha * A_h @ x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((A.shape[0],K))
    y = b.rand_array((A.shape[1],K))
    y_exp = beta * y.to_host() + alpha * A_h.H @ x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (M,N*stack)
    assert A.H.shape == (N*stack,M)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,stack,M,N,K,density,alpha,beta",
    product( BACKENDS, [1,2,3], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1] ))
def test_BlockDiag(backend, stack, M, N, K, density, alpha, beta):
    b = backend()
    mats_h = [indigo.util.randM(M,N,density) for i in range(stack)]
    A_h = spp.block_diag(mats_h)

    mats_d = [b.SpMatrix(m) for m in mats_h]
    A = b.BlockDiag(mats_d)

    # forward
    x = b.rand_array((A.shape[1],K))
    y = b.rand_array((A.shape[0],K))
    y_exp = beta * y.to_host() + alpha * A_h @ x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((A.shape[0],K))
    y = b.rand_array((A.shape[1],K))
    y_exp = beta * y.to_host() + alpha * A_h.H @ x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (M*stack,N*stack)
    assert A.H.shape == (N*stack,M*stack)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,stack,M,N,K,density,alpha,beta",
    product( BACKENDS, [1,2,3], [5,6], [7,8], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1] ))
def test_KronI(backend, stack, M, N, K, density, alpha, beta):
    b = backend()
    mat_h = indigo.util.randM(M,N,density)
    A_h = spp.kron( spp.eye(stack), mat_h )

    mat_d = b.SpMatrix(mat_h)
    A = b.KronI(stack, mat_d)

    # forward
    x = b.rand_array((A.shape[1],K))
    y = b.rand_array((A.shape[0],K))
    y_exp = beta * y.to_host() + alpha * A_h @ x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((A.shape[0],K))
    y = b.rand_array((A.shape[1],K))
    y_exp = beta * y.to_host() + alpha * A_h.H @ x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # shape
    assert A.shape == (M*stack,N*stack)
    assert A.H.shape == (N*stack,M*stack)

    # dtype
    assert A.dtype == np.dtype('complex64')


@pytest.mark.parametrize("backend,M,N,K,B",
    product( BACKENDS, [22,23,24], [22,23,24], [22,23,24], [1,2,3,8] )
)
def test_UnscaledFFT_3d(backend, M, N, K, B ):
    b = backend()

    # forward
    x = b.rand_array( (M*N*K, B) )
    y = b.rand_array( (M*N*K, B) )
    x_h = x.to_host().reshape( (M,N,K,B), order='F' )

    A = b.UnscaledFFT( (M,N,K), dtype=x.dtype )

    A.eval(y, x)

    y_exp = np.fft.fftn( x_h, axes=(0,1,2) )
    y_act = y.to_host().reshape( (M,N,K,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)

    # adjoint
    x = b.rand_array( (M*N*K, B) )
    y = b.rand_array( (M*N*K, B) )
    x_h = x.to_host().reshape( (M,N,K,B), order='F' )

    A.H.eval(y, x)

    y_exp = np.fft.ifftn( x_h, axes=(0,1,2) ) * (M*N*K)
    y_act = y.to_host().reshape( (M,N,K,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)

@pytest.mark.parametrize("backend,M,N,B",
    product( BACKENDS, [22,23,24], [22,23,24], [1,2,3,8] )
)
def test_UnscaledFFT_2d(backend, M, N, B ):
    b = backend()

    # forward
    x = b.rand_array( (M*N, B) )
    y = b.rand_array( (M*N, B) )
    x_h = x.to_host().reshape( (M,N,B), order='F' )

    A = b.UnscaledFFT( (M,N), dtype=x.dtype )

    A.eval(y, x)

    y_exp = np.fft.fftn( x_h, axes=(0,1) )
    y_act = y.to_host().reshape( (M,N,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)

    # adjoint
    x = b.rand_array( (M*N, B) )
    y = b.rand_array( (M*N, B) )
    x_h = x.to_host().reshape( (M,N,B), order='F' )

    A.H.eval(y, x)

    y_exp = np.fft.ifftn( x_h, axes=(0,1) ) * (M*N)
    y_act = y.to_host().reshape( (M,N,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,M,B",
    product( BACKENDS, [22,23,24], [1,2,3,8] )
)
def test_UnscaledFFT_1d(backend, M, B ):
    b = backend()

    # forward
    x = b.rand_array( (M, B) )
    y = b.rand_array( (M, B) )
    x_h = x.to_host().reshape( (M,B), order='F' )

    A = b.UnscaledFFT( (M,), dtype=x.dtype )

    A.eval(y, x)

    y_exp = np.fft.fftn( x_h, axes=(0,) )
    y_act = y.to_host().reshape( (M,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)

    # adjoint
    x = b.rand_array( (M, B) )
    y = b.rand_array( (M, B) )
    x_h = x.to_host().reshape( (M,B), order='F' )

    A.H.eval(y, x)

    y_exp = np.fft.ifftn( x_h, axes=(0,) ) * M
    y_act = y.to_host().reshape( (M,B), order='F' )
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,M,N,K,B",
    product( BACKENDS, [22,23,24], [22,23,24], [22,23,24], [1,2,3,8])
)
def test_UnitaryFFT(backend, M, N, K, B ):
    b = backend()
    A = b.FFT( (M,N,K), dtype=np.dtype('complex64') )

    x = b.rand_array( (M*N*K, B) )
    y = b.rand_array( (M*N*K, B) )
    x_exp = x.to_host().reshape( (M,N,K,B), order='F' )

    A.eval(y, x)
    A.H.eval(x, y)

    x_act = x.to_host().reshape( (M,N,K,B), order='F' )
    npt.assert_allclose(x_act, x_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,M,N,K,B",
    product( BACKENDS, [22,23,24], [22,23,24], [22,23,24], [1,2,3,8])
)
def test_CenteredFFT(backend, M, N, K, B ):
    from numpy.fft import fftshift, ifftshift, fftn, ifftn

    b = backend()
    A = b.FFTc( (M,N,K), dtype=np.dtype('complex64') )

    # forward
    ax = (0,1,2)
    x = b.rand_array( (M*N*K,B) )
    y = b.rand_array( (M*N*K,B) )
    x_h = x.to_host().reshape( (M,N,K,B), order='F' )

    A.eval(y, x)

    y_act = y.to_host().reshape( (M,N,K,B), order='F' )
    y_exp = fftshift( fftn( ifftshift(x_h, axes=ax), axes=ax, norm='ortho'), axes=ax)
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)

    # adjoint
    x = b.rand_array( (M*N*K,B) )
    y = b.rand_array( (M*N*K,B) )
    x_h = x.to_host().reshape( (M,N,K,B), order='F' )

    A.H.eval(y, x)

    y_act = y.to_host().reshape( (M,N,K,B), order='F' )
    y_exp = fftshift( ifftn( ifftshift(x_h, axes=ax), axes=ax, norm='ortho'), axes=ax)
    npt.assert_allclose(y_act, y_exp, rtol=1e-2)


@pytest.mark.parametrize("backend,batch,x,y,z,px,py,pz",
    product( BACKENDS, [1,2,4,8], [3,4],[3,4],[3,4],
                                  [0,1,2],[0,1,2],[0,1,2] )
)
def test_zpad(backend, batch, x, y, z, px, py, pz):
    N = (z, y, x, batch)
    M = (z+2*pz, y+2*py, x+2*px, batch)
    u = np.random.rand(*N) + 1j*np.random.rand(*N)
    v = np.random.rand(*M) + 1j*np.random.rand(*M)
    u = np.require(u, dtype=np.dtype('complex64'), requirements='F')
    v = np.require(v, dtype=np.dtype('complex64'), requirements='F')

    b = backend()
    A = b.Zpad( M[:3], N[:3], dtype=v.dtype )
    
    # check adjoint
    u_d = b.copy_array(u.reshape((-1,batch), order='F'))
    v_d = b.copy_array(v.reshape((-1,batch), order='F'))
    A.H.eval(u_d, v_d)
    u_h = u_d.to_host().reshape(N, order='F')
    v_inner = v[ pz:(-pz or None), py:(-py or None), px:(-px or None) ]
    np.testing.assert_allclose(v_inner, u_h)

    # check forward
    u_d = b.copy_array(u.reshape((-1,batch), order='F'))
    v_d = b.copy_array(v.reshape((-1,batch), order='F'))
    A.eval(v_d, u_d)
    v_h = v_d.to_host().reshape(v.shape, order='F')
    v_inner = v_h[ pz:(-pz or None), py:(-py or None), px:(-px or None) ]
    np.testing.assert_allclose(v_inner, u)


@pytest.mark.parametrize("backend,batch,x,y,z,ro,tr",
    product( BACKENDS, [1,2,4,8], [3,4],[3,4],[3,4],[5,6],[7,8] )
)
def test_interp(backend, batch, x, y, z, ro, tr):
    N = (z, y, x, batch)
    M = (1, ro, tr, batch)
    u = np.random.rand(*N) + 1j*np.random.rand(*N)
    v = np.random.rand(*M) + 1j*np.random.rand(*M)
    u = np.require(u, dtype=np.dtype('complex64'), requirements='F')
    v = np.require(v, dtype=np.dtype('complex64'), requirements='F')

    b = backend()
    width = 3
    table = np.random.random(128)
    coord = np.random.rand(3, *M[1:3])
    A = b.Interp( N[:3], coord, width, table, dtype=u.dtype )

    u_d = b.copy_array(u.reshape((-1,batch), order='F'))
    v_d = b.copy_array(v.reshape((-1,batch), order='F'))
    A.H.eval(u_d, v_d)
    AHv = u_d.to_host().flatten(order='F')

    u_d = b.copy_array(u.reshape((-1,batch), order='F'))
    v_d = b.copy_array(v.reshape((-1,batch), order='F'))
    A.eval(v_d, u_d)
    Au = v_d.to_host().flatten(order='F')

    # adjointness test
    u = u.flatten(order='F')
    v = v.flatten(order='F')
    np.testing.assert_allclose( np.dot(Au,v), np.dot(u,AHv), atol=1e-3 )


@pytest.mark.parametrize("backend,batch,M,N,c1,c2",
    product( BACKENDS, [1,2,4,8], [3,4],[3,4], [4,5],[3,6] )
)
def test_nested_kroni(backend, batch, M, N, c1, c2):
    A00_h = indigo.util.randM(M, N, 0.9)
    A0_h = spp.kron( spp.eye(c1), A00_h )
    A_h = spp.kron( spp.eye(c2), A0_h )

    b = backend()
    A00 = b.SpMatrix(A00_h)
    A0 = b.KronI(c1, A00)
    A = b.KronI(c2, A0)

    V, U = A.shape
    u = np.random.rand(U) + 1j*np.random.rand(U)
    u = np.require(u, dtype=np.dtype('complex64'), requirements='F')
    v = np.random.rand(V) + 1j*np.random.rand(V)
    v = np.require(v, dtype=np.dtype('complex64'), requirements='F')

    # forward
    u_d = b.copy_array(u)
    v_d = b.copy_array(v)
    A.eval(v_d, u_d)
    v_act = v_d.to_host()
    v_exp = A_h @ u
    np.testing.assert_allclose( v_act, v_exp, rtol=1e-6 )

    # adjoint
    u_d = b.copy_array(u)
    v_d = b.copy_array(v)
    A.H.eval(u_d, v_d)
    u_act = u_d.to_host()
    u_exp = A_h.H @ v
    np.testing.assert_allclose( u_act, u_exp, rtol=1e-6 )


@pytest.mark.parametrize("backend,M,N,K,forward,alpha,beta",
    product( BACKENDS, [23,24,45], [45,24,23], [1,8,9,17], [True,False], [0,.5,1], [0,.5,1] ))
def test_DenseMatrix(backend, M, N, K, forward, alpha, beta):
    b = backend()
    A_h = indigo.util.rand64c(M,N)
    A = b.DenseMatrix(A_h)

    if forward:
        x = b.rand_array((N,K))
        y = b.rand_array((M,K))
        y_exp = beta * y.to_host() + alpha * A_h.dot(x.to_host())
        A.eval(y, x, alpha=alpha, beta=beta)
    else:
        x = b.rand_array((M,K))
        y = b.rand_array((N,K))
        y_exp = beta * y.to_host() + alpha * np.conj(A_h.T).dot(x.to_host())
        A.H.eval(y, x, alpha=alpha, beta=beta)

    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)


@pytest.mark.parametrize("backend,M,N,K,density,alpha,beta,batch",
    product( BACKENDS, [23,45], [45,23], [1,8,9,17], [0.01,0.1,0.5,1], [0,.5,1], [0,.5,1],
             [None,1,2,10] ))
def test_batch(backend, M, N, K, density, alpha, beta, batch):
    b = backend()
    A_h = indigo.util.randM(M, N, density)
    A = b.SpMatrix(A_h, batch=batch)

    # forward
    x = b.rand_array((N,K))
    y = b.rand_array((M,K))
    y_exp = beta * y.to_host() + alpha * A_h * x.to_host()
    A.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)

    # adjoint
    x = b.rand_array((M,K))
    y = b.rand_array((N,K))
    y_exp = beta * y.to_host() + alpha * A_h.H * x.to_host()
    A.H.eval(y, x, alpha=alpha, beta=beta)
    npt.assert_allclose(y.to_host(), y_exp, rtol=1e-5)
